from .Button import Button
from .Slider import Slider
from .ColorPicker import ColorPicker
from .DatePicker import DatePicker
from .Progress import Progress
from .Radio import Radio
# from .Menu import Menu

from .deprecated import button
from .deprecated import slider
from .deprecated import colorPicker
from .deprecated import datePicker
from .deprecated import progress
from .deprecated import radio
# from .deprecated import menu


"""The switch widget provides users the ability to choose between two distinct values."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
except:
    import settings

    
#####################################################################################################################################################
# Switch control
#####################################################################################################################################################
class switch():
    """
    The switch widget provides users the ability to choose between two distinct values.

    Parameters
    ----------
    flag : int
        Initial value of the switch
    label : str
        Text to display on the left of the switch widget
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    inset : bool, optional
        Flag to enlarge switch track to encompass the thumb (default is True)
    dense : bool, optional
        If True, the widget is displayed with smaller dimensions (default is False)
    onchange : function, optional
        Python function to call when the user clicks on the switch. The function will receive a parameter of type bool containing the status of the switch flag
            
    Example
    -------
    Creation and display of a switch widget::
        
        from vois.vuetify import switch
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def onchange(value):
            with output:
                print(value)

        s = switch.switch(True, "Label of the switch", inset=True, onchange=onchange)

        display(s.draw())
        display(output)

    .. figure:: figures/switch.png
       :scale: 100 %
       :alt: switch widget

       Switch widget with inset flag True.
   """

    # Initialization
    def __init__(self, flag, label, color=settings.color_first, inset=True, dense=False, onchange=None):
        
        self.onchange = onchange
        
        self.switch = v.Switch(v_model=bool(flag), dense=dense, flat=True, label=label, color=color, inset=inset, class_="pa-0 ma-0 ml-3 mt-2 mb-n3", disabled=False)
        
        # If requested onchange management
        if not self.onchange is None:
            self.switch.on_event('change', self.__internal_onchange)
    
    
    # Get the value
    @property
    def value(self):
        """
        Get/Set the status of the switch.
        
        Returns
        --------
        flag : bool
            Status of the switch

        Example
        -------
        Programmatically set the switch status and print it::
            
            t.value = True
            print(t.value)
        
        """
        return self.switch.v_model
    
    
    # Set the value of the switch
    @value.setter
    def value(self, flag):
        self.switch.v_model = bool(flag)
        if self.onchange:
            self.onchange(self.switch.v_model)
    
    
    # Manage onchange event
    def __internal_onchange(self, widget=None, event=None, data=None):
        if self.onchange:
            self.onchange(self.switch.v_model)
    
    # Returns the vuetify object to display (the v.Container)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Html that has a v.Switch widget as its only child)"""
        return v.Html(tag='div',children=[self.switch], style_='overflow: hidden;')
            

    # disabled property
    @property
    def disabled(self):
        """
        Get/Set the disabled state.
        """
        return self.switch.disabled
    
    @disabled.setter
    def disabled(self, flag):
        self.switch.disabled = flag        


"""Input widget to select a color"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
from ipywidgets import widgets
from vois import colors
from vois.vuetify import popup
from vois.vuetify.utils.util import *
from typing import Callable, Any, Optional
import warnings


# Popup widget that, on hover, opens a card with complementary, triadic, analogous, etc. clickable colors
class colorTheoryPopup():

    # Initialisation
    def __init__(self, picker):
        
        self.picker = picker
        self._color = self.picker.color

        card_width  = 564
        card_height = 320

        self.card  = v.Card(flat=True, width=card_width, height=card_height, class_='pa-2 ma-0')
        self.popup = popup.popup(self.card, '', text=False, outlined=False, icon='mdi-palette', color='#333333', rounded=False,
                                 buttonwidth=30, buttonheight=self.picker.height+1, popupwidth=card_width, popupheight=card_height)
        self.updateCard()
        
    
    # Update the content of the card to be dispalyed when hover on the popup
    def updateCard(self):
        l_title = self.label('Color theory helpers:', width=300, weight=550)
        
        l_compl = self.label('Complementary colors:', width=160)
        c_col   = self.colorcard(self._color)
        c_compl = self.colorcard(colors.rgb2hex(colors.complementaryColor(colors.string2rgb(self._color))))

        l_tri = self.label('Triadic colors:', width=160)
        tri = [colors.rgb2hex(x) for x in colors.triadicColor(colors.string2rgb(self._color))]
        c_tri1 = self.colorcard(tri[0])
        c_tri2 = self.colorcard(tri[1])

        l_split = self.label('Split complementary col.:', width=160)
        split = [colors.rgb2hex(x) for x in colors.splitComplementaryColor(colors.string2rgb(self._color))]
        c_split1 = self.colorcard(split[0])
        c_split2 = self.colorcard(split[1])

        l_tet = self.label('Tetradic colors:', width=160)
        tet = [colors.rgb2hex(x) for x in colors.tetradicColor(colors.string2rgb(self._color))]
        c_tet1 = self.colorcard(tet[0])
        c_tet2 = self.colorcard(tet[1])
        c_tet3 = self.colorcard(tet[2])
        c_tet4 = self.colorcard(tet[3])

        l_squ = self.label('Square colors:', width=160)
        squ = [colors.rgb2hex(x) for x in colors.squareColor(colors.string2rgb(self._color))]
        c_squ1 = self.colorcard(squ[0])
        c_squ2 = self.colorcard(squ[1])
        c_squ3 = self.colorcard(squ[2])
        
        l_ana = self.label('Analogous colors:', width=160)
        ana = [colors.rgb2hex(x) for x in colors.analogousColor(colors.string2rgb(self._color))]
        c_ana1 = self.colorcard(ana[0])
        c_ana2 = self.colorcard(ana[1])

        l_mono = self.label('Monochromatic colors:', width=160)
        c_mono1 = self.colorcard(colors.rgb2hex(colors.monochromaticColor(colors.string2rgb(self._color), increment=-0.6)))
        c_mono2 = self.colorcard(colors.rgb2hex(colors.monochromaticColor(colors.string2rgb(self._color), increment=-0.4)))
        c_mono3 = self.colorcard(colors.rgb2hex(colors.monochromaticColor(colors.string2rgb(self._color), increment=-0.2)))
        c_mono4 = self.colorcard(colors.rgb2hex(colors.monochromaticColor(colors.string2rgb(self._color), increment= 0.2)))
        c_mono5 = self.colorcard(colors.rgb2hex(colors.monochromaticColor(colors.string2rgb(self._color), increment= 0.4)))
        c_mono6 = self.colorcard(colors.rgb2hex(colors.monochromaticColor(colors.string2rgb(self._color), increment= 0.6)))
        
        w1 = widgets.VBox([
                l_title,
                widgets.HBox([l_compl, c_col, c_compl]),
                widgets.HBox([l_tri,   c_col, c_tri1, c_tri2]),
                widgets.HBox([l_split, c_col, c_split1, c_split2]),
        ])
            
        self.card.children = [
            widgets.VBox([
                widgets.HBox([w1, v.Html(tag='div', style_='width: 84px;'), v.Img(src=colors.colorWheel, width=130, height=130)]),
                widgets.HBox([l_tet,   c_col, c_tet1, c_tet2, c_tet3, c_tet4]),
                widgets.HBox([l_squ,   c_col, c_squ1, c_squ2, c_squ3]),
                widgets.HBox([l_ana,   c_col, c_ana1, c_ana2]),
                widgets.HBox([l_mono,  c_mono1, c_mono2, c_mono3, c_col, c_mono4, c_mono5, c_mono6])
            ])
        ]
        

    # Returns the vuetify object to display
    def draw(self):
        return self.popup.draw()
        
        
    # color property
    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, c):
        if isinstance(c, str):
            self._color = c
            self.updateCard()
        
        
    # disabled property
    @property
    def disabled(self):
        return self.popup.disabled
    
    @disabled.setter
    def disabled(self, flag):
        self.popup.disabled = flag
        
        
    # Creation of a label
    def label(self, text, class_='pa-0 ma-0 mr-3 mb-4', size=14, weight=400, color='#000000', width=None):
        lab = v.Html(tag='div', children=[text], class_=class_)
        if width is None: lab.style_ = 'font-size: %dpx; font-weight: %d; color: %s; overflow: hidden;'%(size,weight,color)
        else:             lab.style_ = 'font-size: %dpx; font-weight: %d; color: %s;  overflow: hidden; width: %dpx'%(size,weight,color,int(width))
        return lab

    
    # Create and returns a card displaying a color and clickable
    def colorcard(self, color, width=54, height=30):

        def onclick(*args):
            self.picker.color = color

        c = v.Card(flat=True, hover=True, color=color, width=width, min_width=width, max_width=width, height=height, tile=True)
        c.on_event('click',onclick)
        return c


    
class ColorPicker(v.Menu):
    """
    Input widget to select a color.

    Parameters
    ----------
    color : str, optional
        Initial color selected on the widget expressed in hexadecimal format '#RRGGBB' (default is '#FF0000')
    dark : bool, optional
        If True, the popup color selection will have a dark background (default is settings.dark_mode)
    dark_text : bool, optional
        If True, the text on the colored button is displayed in white, if False in black. If dark_text is None, the color of the text is automatically selected based on the currently selected color (default is None)
    width : int, optional
        Width of the widget in pixels (default is 40)
    height : int, optional
        Height of the widget in pixels (default is 30)
    rounded : bool, optional
        If True the color widget is displayed as a round button (default is False)
    canvas_height : int, optional
        Height of the canvas displayed on top of the popup window to select the colors (default is True)
    show_canvas : bool, optional
        If True the popup window will show the color canvas (default is True)
    show_mode_switch : bool, optional
        If True the popup window will show mode switch control among RGB, HSL and HAX (default is True)
    show_inputs : bool, optional
        If True the popup window will show the input field for the color components (default is True)
    show_swatches : bool, optional
        If True the popup window will show the color swatches (default is True)
    swatches_max_height : int, optional
        Height in pixels of the swatches area in the popup window (default is 164)
    text : str, optional
        Text to display in the color button (default is '')
    text_weight : int, optional
        Weight of the text to be shown in the button (default is 400, Bold is any value greater or equal to 500)
    on_change : function, optional
        Python function to call when the user selects a different color. The function will receive the argument parameter. If the argument is None, the function will receive no parameters. (default is None)
    argument : any, optional
        Argument to be passed to the onchange function (default is None)
    offset_x : bool, optional
        If True the popup window will be opened on the right of the color button (default is False)
    offset_y : bool, optional
        If True the popup window will be opened on the bottom of the color button (default is True)
    disabled : bool, optional
        True if the selection of the color is disabled, False otherwise (default is False)
    color_theory_popup : bool, optional
        If True a popup window (self.ctpopup) is created to show the complementary, analogous, etc.. colors on hover (default is False)

    Example
    -------
    Creation of a color picker widget to select a color::
        
        from vois.vuetify import ColorPicker
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def on_change():
            with output:
                print('Changed to', c.color)

        c = ColorPicker(color='#00AAFF',
                        width=30, height=30,
                        rounded=False,
                        on_change=on_change,
                        offset_x=True,
                        offset_y=False)

        display(c)
        display(output)

    .. figure:: figures/colorPicker.png
       :scale: 100 %
       :alt: colorPicker widget

       Example of a colorPicker to select a color
    """

    def hex2rgb(self, color):
        if color[0] == '#':
            color = color[1:]
        rgb = (int(color[0:2], 16), int(color[2:4], 16), int(color[4:6], 16))
        return rgb

    def textDark(self, color):
        r, g, b = self.hex2rgb(color)
        if r + g + b <= 255 + 128:
            return True
        return False

    deprecation_alias = dict(textweight='text_weight', onchange='on_change')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 color: str = "#FF0000",
                 dark: bool = None,
                 dark_text: bool = None,
                 width: int = 40,
                 height: int = 30,
                 rounded: bool = False,
                 canvas_height: int = 100,
                 show_canvas: bool = True,
                 show_mode_switch: bool = True,
                 show_inputs: bool = True,
                 show_swatches: bool = True,
                 swatches_max_height: int = 164,
                 text: str = '',
                 text_weight: int = 400,
                 on_change: Optional[Callable[[], None]] = None,
                 argument: Optional[Any] = None,
                 offset_x: bool = False,
                 offset_y: bool = True,
                 disabled: bool = False,
                 color_theory_popup: bool = False):

        from vois.vuetify import settings

        self._color = str(color).upper()
        self.dark = dark if dark is not None else settings.dark_mode
        self._dark_text = dark_text
        self.width = width
        self.height = height
        self.rounded = rounded
        self.canvas_height = canvas_height
        self.show_canvas = show_canvas
        self.show_mode_switch = show_mode_switch
        self.show_inputs = show_inputs
        self.show_swatches = show_swatches
        self.swatches_max_height = swatches_max_height
        self.text = text
        self.text_weight = text_weight
        self.on_change = on_change
        self.argument = argument
        self.offset_x = offset_x
        self.offset_y = offset_y
        self._disabled = disabled
        
        self.ctpopup = None
        if color_theory_popup:
            self.ctpopup = colorTheoryPopup(self)

        if self._disabled:
            von = ''
        else:
            von = 'menuData.on'
            
        if self._dark_text is None:
            darktext = self.textDark(self._color)
        else:
            darktext = self._dark_text
            
        self.button = v.Btn(v_on=von, depressed=True, large=False, dense=True, class_='pa-0 ma-0', color=self._color,
                            rounded=self.rounded, height=self.height, width=self.width, min_width=self.width, dark=darktext,
                            style_='font-weight: %d; text-transform: none' % self.text_weight, children=[self.text])

        self.p = v.ColorPicker(value=self._color, flat=True, class_="pa-0 ma-0", style_="min-width: 300px;",
                               canvas_height=self.canvas_height, show_swatches=self.show_swatches, dark=self.dark,
                               swatches_max_height=self.swatches_max_height, hide_canvas=not self.show_canvas,
                               hide_mode_switch=not self.show_mode_switch, hide_inputs=not self.show_inputs)
        self.p.on_event('input', self.__internal_onchange)

        super().__init__(offset_x=self.offset_x, offset_y=self.offset_y, open_on_hover=False, dense=True,
                         close_on_click=True, close_on_content_click=False,
                         v_slots=[{'name': 'activator', 'variable': 'menuData', 'children': self.button}],
                         children=[self.p])

        self.children = [self.p]

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Manage 'input' event
    def __internal_onchange(self, widget, event, data):
        if data.upper() != self._color.upper():
            self._color = data.upper()
            self.button.color = self._color
            if self.ctpopup is not None:
                self.ctpopup.color = self._color
            if self._dark_text is None:
                self.button.dark = self.textDark(self._color)
            if self.on_change:
                if self.argument is None:
                    self.on_change()
                else:
                    self.on_change(self.argument)

    # Returns the vuetify object to display (the v.Menu)
    def draw(self):
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self

    # color property
    @property
    def color(self):
        """
        Get/Set the selected color.
        
        Returns
        --------
        c : str
            color currently selected

        Example
        -------
        Programmatically change the color::
            
            picker.color = '#00FF00'
            print(picker.color)
        
        """
        return self._color

    @color.setter
    def color(self, c):
        if isinstance(c, str):
            self._color = c
            self.p.value = self._color
            self.button.color = self._color
            if self.ctpopup is not None:
                self.ctpopup.color = self._color
            if self._dark_text is None:
                self.button.dark = self.textDark(self._color)
            if self.on_change:
                if self.argument is None:
                    self.on_change()
                else:
                    self.on_change(self.argument)

    # disabled property
    @property
    def disabled(self):
        """
        Get/Set the disabled state of the widget.
        
        Returns
        --------
        flag : bool
            True if the widget is disabled, False otherwise

        Example
        -------
        Programmatically change the disabled state::
            
            picker.disabled = True
            print(picker.disabled)
        
        """
        return self._disabled

    @disabled.setter
    def disabled(self, flag):
        self._disabled = bool(flag)
        if self._disabled:
            von = ''
            self.button.children = ['⊗']   # Visual change to let the user know that the widget is disabled
        else:
            von = 'menuData.on'
            self.button.children = [self.text]
        self.button.v_on = von
        
        if self.ctpopup is not None:
            self.ctpopup.disabled = self._disabled

        
    # dark_text property
    @property
    def dark_text(self):
        """
        Get/Set the dark flag for the button that displayes the selected color. If True, the text on the colored button is displayed in white, if False in black. If dark_text is None, the color of the text is automatically selected based on the currently selected color
        
        Returns
        --------
        flag : bool
            If True, the text on the colored button is displayed in white, if False in black. If dark_text is None, the color of the text is automatically selected based on the currently selected color

        Example
        -------
        Programmatically change the dark_text property::
            
            picker.dark_text = True
            print(picker.dark_text)
        
        """
        return self._dark_text

    @dark_text.setter
    def dark_text(self, flag):
        self._dark_text = bool(flag)
        
        if self._dark_text is None:
            darktext = self.textDark(self._color)
        else:
            darktext = self._dark_text
        self.button.dark = darktext
        


"""Utility functions for the creation ipytrees from hierarchical data."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from traitlets import Float
from ipytree import Tree, Node
from IPython.display import display
import pandas as pd
import numpy as np


##################################################################################################################################
# Creation of ipytree
##################################################################################################################################

# Node of the tree that stores a float value (Very slow!!!)
class DataNode(Node):
    value = Float(0.0)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name
    
# Handle click on a node
def basic_handle_click(event):
    if event['new']:
        print(event.owner.value)

        
        
##################################################################################################################################
# Create a ipytree from a list of names with an implicit tree structure, example: ['JRC', 'JRC.D', 'JRC.D.3', ...]
# EXAMPLE:
# def pippo(event):
#     if event['new']:
#         print(event.owner.value)
#         
# tree, nodes, parent_of = createIpytreeFromList(['A','A.1','A.2','A.1.1'], rootName='A', valuefor={'A.1.1': 10.0}, handle_click=pippo)
# display(tree)
##################################################################################################################################
def createIpytreeFromList(nameslist=[], rootName='', separator='.', valuefor={}, handle_click=basic_handle_click, select_root=False):
    """
    Create a ipytree from a list of names with an implicit tree structure, example: ['JRC', 'JRC.D', 'JRC.D.3', ...]
    
    Parameters
    ----------
        nameslist : list of strings, optional
            List of strings that contain a hierarchical structure, considering the separator character (default is [])
        rootName: str, optional
            Name to be displayed as root of the tree (default is '')
        separator: str, optional
            String or character to be considered as separator for extracting the hierarchical structure from the nameslist list of strings (default is '.')
        valuefor : dict, optional
            Dictionary to assign a numerical value to each node of the tree (default is {})
        handle_click : function, optional
            Python function to call when the selected nodes change caused by user clicking (default is ipytree.basic_handle_click)
        select_root : bool, optional
            If True the root node is selected at start (default is False)
        
    Returns
    -------
        A tuple of 3 elements: the tree instance, a dict containing the info on the nodes, a dict containing the parent of each of the nodes
    
    
    Example
    -------
    Example of the creation of a ipytree from a list of strings with hierarchy defined by the '.' character::
    
        from vois import ipytrees
        
        def onclick(event):
            if event['new']:
                print(event.owner, event.owner.value)

        tree, n, p = ipytrees.createIpytreeFromList(['A','A.1','A.2',
                                                    'A.1.1','A.3.1',
                                                    'A.4.1.2','A.5.2.1',
                                                    'A.4.2.3.1','B','B.A'],
                                                    rootName='Directorates',
                                                    valuefor={'A.1.1': 10.0,
                                                              'A.3.1': 5.0},
                                                    handle_click=onclick)
        display(tree)
        
    .. figure:: figures/ipytree.png
       :scale: 100 %
       :alt: ipytrees example

       Ipytree produced by the example code
       
    """
    
    # Add a node to a tree-node
    def addNode(name, value, parent):
        nonlocal nodes, parent_of
        if not name in nodes:
            elem = DataNode(name, value=value)
            elem.open_icon_style  = 'info'
            elem.close_icon_style = 'info'
            elem.show_icon = False
            elem.opened = False
            elem.observe(handle_click, 'selected')
            parent.add_node(elem)
            nodes[name] = elem
            parent_of[name] = parent
            while parent != tree:
                parent.value += value
                parent = parent_of[parent.name]
            return elem
    
    nodes     = {}
    parent_of = {}
    tree = Tree(stripes=False)
    root = tree
    if len(rootName) > 0:
        root = addNode(rootName,0.0,tree)

    # Manage an implicit tree with separator!
    for name in nameslist:
        parts = name.split(separator)
        fullname = ''
        for part in parts:
            parentname = fullname
            if len(fullname) == 0: fullname += part
            else:                  fullname += '.' + part

            parent = root
            if parentname in nodes: parent = nodes[parentname]

            value = 0.0
            if fullname in valuefor: value = valuefor[fullname]
            addNode(fullname,value,parent)

    if rootName in nodes:            
        root = nodes[rootName]
        root.opened = True
        
        if select_root:
            root.selected = True
        
    return tree, nodes, parent_of
    
    
    
##################################################################################################################################
# Create a ipytree from two columns of a Dataframe. An optional third column containing a numeric value can be used
##################################################################################################################################
def createIpytreeFromDF2Columns(df, colindexLabels1, colindexLabels2, colindexValues=-1, rootName='', handle_click=basic_handle_click, select_root=False):
    """
    Create a two level ipytree from two columns of a Pandas DataFrame
    
    Parameters
    ----------
        df : Pandas DataFrame
            Input Pandas DataFrame
        colindexLabels1 : int
            Index of the column that contains the labels of the first level of the tree
        colindexLabels2 : int
            Index of the column that contains the labels of the second level of the tree
        colindexValues : int, optional
            Index of the column that contains the values for the nodes of the tree
        rootName: str, optional
            Name to be displayed as root of the tree (default is '')
        handle_click : function, optional
            Python function to call when the selected nodes change caused by user clicking (default is ipytree.basic_handle_click)
        select_root : bool, optional
            If True the root node is selected at start (default is False)
        
    Returns
    -------
        A tuple of 3 elements: the tree instance, a dict containing the info on the nodes, a dict containing the parent of each of the nodes
        
    """

    # Add a node to a tree-node
    def addNode(name, value, parent):
        nonlocal nodes, parent_of
        if not name in nodes:
            elem = DataNode(name, value=value)
            elem.open_icon_style  = 'info'
            elem.close_icon_style = 'info'
            elem.show_icon = False
            elem.opened = False
            elem.observe(handle_click, 'selected')
            parent.add_node(elem)
            nodes[name] = elem
            parent_of[name] = parent
            while parent != tree:
                parent.value += value
                parent = parent_of[parent.name]
            return elem
    
    nodes     = {}
    parent_of = {}
    tree = Tree(stripes=False)
    root = tree
    if len(rootName) > 0:
        root = addNode(rootName,0.0,tree)

    # Create the tree from two columns of a df
    for index,row in df.iterrows():
        parentname = str(row[colindexLabels1])
        name       = str(row[colindexLabels2])
        value = 0.0
        if colindexValues >= 0: value = float(row[colindexValues])

        addNode(parentname, 0.0, root)
        parent = nodes[parentname]
        addNode(name,value,parent)

    root.opened = True
    if select_root:
        root.selected = True

    return tree, nodes, parent_of






"""Multiple selection widget."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import fontsettings
except:
    import settings
    import fontsettings



#####################################################################################################################################################
# Class to display a multiple selection widget
#####################################################################################################################################################
class selectMultiple():
    """
    Single selection widget from a dropdown list. Passing the parameter newvalues_enabled=True enables the user to insert new strings in the widget.
        
    Parameters
    ----------
    label : str
        Help text to display 
    values : list of strings
        Strings to be displayed in the dropdown list of the widget 
    selected : list of str, optional
        List of strings that are initially selected (default is [])
    mapping : function, optional
        Python function to call to transform the visible strings into codes (for example names of countries to their iso2 codes)
    reverse_mapping : function, optional
        Python function to call to transform the codes into visible strings (for example iso2 codes of countries to their names)
    onchange : function, optional
        Python function to call when the user selects one of the values in the list. The function will receive no parameter (use value property to retrieve the current selection)
    width : int, optional
        Width in pixel of the widget (default is 300 pixels)
    marginy : int, optional
        Margin in y coordinates to position the widget from top (default is 1)
            
    Example
    -------
    Creation and display of a multiple selection widget for the selection of one or more countries::
        
        from vois.vuetify import selectMultiple
        from ipywidgets import widgets
        from IPython.display import display
        
        output = widgets.Output()
        display(output)

        def onchange():
            with output:
                print(sel.value)

        sel = selectMultiple.selectMultiple('Country:',
                                            ['Belgium', 'France', 'Italy', 'Germany'],
                                            selected=['France', 'Italy'],
                                            width=200,
                                            onchange=onchange)
        sel.draw()

    .. figure:: figures/selectMultiple.png
       :scale: 100 %
       :alt: selectMultiple widget

       Example of a selectMultiple widget to select from a list of countries.
   """
    
    # Initialization
    def __init__(self, label, values, selected=[], mapping=None, reverse_mapping=None, width=300, onchange=None, marginy=1):
        self.label           = label
        self.values          = values
        self.mapping         = mapping           # Function to convert names to codes
        self.reverse_mapping = reverse_mapping   # Function to convert codes to names
        self.width           = width
        self.onchange        = onchange
        
        self.select = None
        self.value  = selected
        
        self.select = v.Select(v_model=self._value,
                               label=self.label, dense=True, solo=False, outlined=True, multiple=True, chips=False, clearable=True,
                               item_color=settings.color_first, color=settings.color_first, class_='pa-0 mx-0 my-%d mb-n6' % marginy,
                               items=self.values, style_='max-width: %dpx; font-family: %s; font-weight:400; text-transform: none' % (self.width, fontsettings.font_name))
        
        self.select.on_event('input', self.__internal_onchange)

        
    def __internal_onchange(self, widget, event, data):
        self._value = data
        if self.onchange: self.onchange()
        
        
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Html widget that contains a v.Select as its unique child)"""
        self.select.value = self._value
        return v.Html(tag='div',children=[self.select], style_='overflow: hidden;')
    
    
    # value property is a list!
    @property
    def value(self):
        """
        Get/Set the selected value.
        
        Returns
        --------
        v : str
            list of strings of the items currently selected

        Example
        -------
        Programmatically select a list of values::
            
            sel.value = ['Italy', 'Belgium']
            print(sel.value)
        
        """
        if not self.mapping is None:
            return [self.mapping(x) for x in self._value]
        return self._value
    

    @value.setter
    def value(self, v):
        if not self.reverse_mapping is None:
            self._value = [self.reverse_mapping(x) for x in v]
        else:
            self._value = v

        if not self.select is None:
            self.select.v_model = self._value
            if self.onchange: self.onchange()

        



"""Utility functions to open a web page."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.core.display import HTML as ipyhtml   # V. I. for urlOpen to work!!!

# Open a web page in another tab
def urlOpen(url, output, target='_blank'):
    """
    Opean a web page in another tab of the browser
    
    Parameters
    ----------
    url : str
        URL of the page to open
    output : instance of ipywidgets.Output() class
        Output widget where the javascript code to open the new page is executed
    target : str, optional
        Target of the open operation (default is '_blank' which means that the page will be opened in a new browser's tab)
        
    Example
    -------
    Open Google page in another tab of the browser::
    
        from vois import urlOpen
        from ipywidgets import widgets, Layout
        from IPython.display import display

        output = widgets.Output(layout=Layout(width='0px', height='0px'))
        display(output)
        
        urlOpen.urlOpen('https://www.google.com', output, target='_blank')
    
        
    Note
    ----
        If the dashboard is created using the :class:`app.app` class, it is preferable to use the function :func:`app.app.urlOpen` that doesn't need the output parameter and uses the Output widget created inside the app instance itself (and invisible to the users)
    
    """
    
    js = '''
<script type=\"text/javascript\">
window.open("%s", "%s");
</script>
''' % (url, target)
        
    with output:
        display(ipyhtml(js))



"""Slider to select a range of numeric values."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
except:
    import settings


#####################################################################################################################################################
# Integer range slider class
#####################################################################################################################################################
class rangeSlider:
    """
    Slider to select a range of numeric values.
        
    Parameters
    ----------
    selectedminvalue : numeric
        Minimum value of the selected range
    selectedmaxvalue : numeric
        Maximum value of the selected range
    minvalue : numeric
        Minimum value selectable by the user
    maxvalue : numeric
        Maximum value selectable by the user
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    onchange : function, optional
        Python function to call when the user selects a value. The function will receive two parameters of numeric type containing the current min and max value selected by the user
    height : int, optional
        Height of the slider widget in pixel (default is 250 pixels)
    vertical : bool, optional
        Flag to display the range slider in vertical mode (default is True)
            
    Example
    -------
    Creation and display of a range slider widget::
        
        from vois.vuetify import rangeSlider
        
        s = rangeSlider.rangeSlider(5,18, 0,20)
        s.draw()

    .. figure:: figures/rangeSlider.png
       :scale: 100 %
       :alt: rangeSlider widget

       RangeSlider widget example
   """

   
    # Initialization
    def __init__(self, selectedminvalue, selectedmaxvalue, minvalue, maxvalue, color=settings.color_first, onchange=None,
                 height=250, vertical=True):
        
        self.selectedminvalue = selectedminvalue
        self.selectedmaxvalue = selectedmaxvalue
        self.minvalue = minvalue
        self.maxvalue = maxvalue
        self.color    = color
        self.onchange = onchange
        self.height   = height
        self.vertical = vertical
        
        if self.vertical: margins = "ml-n5 mr-1 mt-n6 mb-n7"
        else:             margins = "ml-5  mr-5 mt-4  mb-n6"
        
        self.slider = v.RangeSlider(v_model=[self.selectedminvalue,self.selectedmaxvalue], dense=True, small=True, thumb_color=self.color, 
                                    thumb_label="always", thumb_size=32, ticks=True, ticks_size=10, 
                                    color=self.color, track_color="grey", class_="pa-0 ma-0 %s" % margins,
                                    min=self.minvalue, max=self.maxvalue, vertical=self.vertical, height=self.height) 
        
        # If requested onchange management
        if not self.onchange is None:
            self.slider.on_event('end', self.__internal_onchange)
        
    
    # Manage onchange event
    def __internal_onchange(self, widget=None, event=None, data=None):
        if self.onchange:
            self.onchange(data[0],data[1])
    
    # Returns the vuetify object to display (the v.Container)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Html widget that contains a v.RangeSlider as its unique child)"""
        return v.Html(tag='div',children=[self.slider], style_='overflow: hidden;')
            



"""Slider widget is a better visualization of the number input. It is used for gathering numerical user data."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

import ipyvuetify as v
from vois.vuetify.utils.util import *
from vois.vuetify import settings
from typing import Callable, Any, Union, Optional


#####################################################################################################################################################
# Integer slider class
#####################################################################################################################################################
class Slider(v.Html):
    """
    Slider widget is a better visualization of the number input. It is used for gathering numerical user data.
        
    Parameters
    ----------
    selected_value : numeric
        Initial value of the slider
    min_value : numeric
        Minimal value selectable by the user
    max_value : numeric
        Maximum value selectable by the user
    step : numeric, optional
        Step interval for ticks (default is 1.0)
    vertical : bool, optional
        Flag that controls the direction of the slider: horizontal or vertical (default is False)
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    on_change : function, optional
        Python function to call when the user selects a value. The function will receive a parameter of numeric type containing the current value of the slider widget
    height : int, optional
        Height of the slider widget in pixel (default is 120 pixels)
    width : int, optional
        Width of the slider widget in pixel. It is needed only for vertical sliders (default is None)
            
    Example
    -------
    Creation and display of a slider widget::
        
        from vois.vuetify import Slider
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()
        display(output)

        def onchange(value):
            with output:
                print(value)

        s = Slider(2015, 2010, 2021, onchange=onchange)
        display(s)

    .. figure:: figures/slider.png
       :scale: 100 %
       :alt: slider widget

       Slider widget example
   """
    deprecation_alias = dict(selectedvalue='selected_value', minvalue='min_value', maxvalue='max_value',
                             onchange='on_change')

    # Initialization
    @deprecated_init_alias(**deprecation_alias)
    def __init__(self,
                 selected_value: Union[int, float],
                 min_value: Union[int, float],
                 max_value: Union[int, float],
                 vertical=False,
                 color: str = settings.color_first,
                 on_change: Optional[Callable[[], None]] = None,
                 height: int = 120,
                 width: int = None,
                 step: Union[int, float] = 1.0,
                 **kwargs):

        super().__init__(**kwargs)

        self.min_value = min_value
        self.max_value = max_value
        self.step = step
        self.vertical = vertical
        self.width = width
        self.on_change = on_change

        if vertical:
            c = "pa-0 ma-0 ml-n12 mr-n12 mt-n9 mb-n10"
        else:
            c = "pa-0 ma-0 ml-5 mr-5 mt-n2 mb-n12"
        self.slider = v.Slider(v_model=selected_value, dense=True, small=True, thumb_color=color,
                               thumb_label="always", thumb_size=32, ticks=True, ticks_size=10,
                               color=color, track_color="grey", class_=c,
                               min=self.min_value, max=self.max_value, step=self.step,
                               vertical=self.vertical, height=height)

        # If requested onchange management
        if not self.on_change is None:
            self.slider.on_event('end', self.__internal_onchange)

        self.tag = 'div'
        self.children = [self.slider]

        if self.vertical and not self.width is None:
            self.style_ = 'width: %dpx; overflow: hidden;' % self.width
        else:
            self.style_ = 'overflow: hidden;'

        for alias, new in self.deprecation_alias.items():
            create_deprecated_alias(self, alias, new)

    # Get the value
    @property
    def value(self):
        """
        Get/Set the current value.
        
        Returns
        --------
        value : numeric
            Current value selected

        Example
        -------
        Programmatically set the value and print it::
            
            s.value = 2012
            print(s.value)
        
        """
        return self.slider.v_model

    # Set the value
    @value.setter
    def value(self, v):
        if self.min_value <= v <= self.max_value:
            self.slider.v_model = v
            if self.on_change:
                self.on_change(self.slider.v_model)

    # Manage onchange event
    def __internal_onchange(self, widget=None, event=None, data=None):
        if self.on_change:
            self.on_change(data)

    # Returns the vuetify object to display (the v.Container)
    def draw(self):
        warnings.warn('The "draw" method is deprecated, please just use the object widget itself.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self



"""Display and selection of a list of SVG files"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
import traitlets
import os
import base64


# Given a path of an SVG file, it reads the content and returns a string containing a IMG tab with the SVG embedded in base64 format
def readSVGbase64(filepath, fill=None, stroke=None):
    svgtext = ''
    with open(filepath,'r') as f:
        svgtext = f.read()
        
    if (not fill is None) and (not stroke is None):
        svgtext.replace('<svg ','<svg fill="%s" stroke="%s"'%(fill,stroke))
    elif not fill is None:
        svgtext.replace('<svg','<svg fill="%s"'%fill)
    elif not stroke is None:
        svgtext.replace('<svg','<svg stroke="%s"'%stroke)

    base64_bytes  = base64.b64encode(bytes(svgtext,'utf-8'))
    base64_string = base64_bytes.decode("ascii")
    return "data:image/svg+xml;base64, %s" % base64_string


# Given a path of an SVG file, it reads the content and returns a string containing a IMG tab with the SVG embedded in base64 format
def readSVGimg(filepath, size=100):
    imgurl = readSVGbase64(filepath)
    return '<img src="%s", style="width:%dpx; height:%dpx; display: flex; justify-content: center; align-items: center; overflow: hidden;">'%(imgurl,size,size) + '</img>'


class svgsGrid(v.VuetifyTemplate):
    """
    Display of a list of SVG files in rows and columns, with possibility to select one of the SVGs.
        
    Parameters
    ----------
    filepaths : list of str
        File paths of the SVG files to display in the grid
    width : str, optional
        Width to use for the display of the SVGs (default is '80px')
    height : str, optional
        Heigth to use for the display of the SVGs (default is '100px' to accomodate for the title of the files)
    cols : int, optional
        Horizontal column span [1,12] for each of the SVGs (default is 1, meaning 12 files are displayed in every row)
    color : str, optional
        Background color of the SVGs (default is 'white')
    ripple : bool, optional
        If True the click on the SVG files is highlighted (default is False)
    svgsize : int, optional
        Size in pixel of the square area where the SVG is displayed (default is 80)
    on_click : function, optional
        Python function to call when the user clicks on one of the SVG files. The function will receive as parameter the index of the clicked SVG. (default is None)

    Example
    -------
    Creation of a grid to display and select SVG files read from a subfolder::
        
        from vois.vuetify import svgsGrid
        from ipywidgets import widgets
        from IPython.display import display
        import glob

        output = widgets.Output()

        def on_click(index):
            with output:
                print(index)

        filepaths = sorted(glob.glob("./maki/icons/*"))

        svgsize = 80
        height  = '%dpx'%(svgsize + 30)
        g = svgsGrid(filepaths=filepaths, cols=1, svgsize=svgsize,
                     height=height, ripple=True, on_click=on_click)

        display(g)
        display(output)

    .. figure:: figures/svgsGrid.png
       :scale: 100 %
       :alt: svgsGrid widget

       Example of an svgsGrid to display a list of SVG files allowing selection
    """
    
    svgitems  = traitlets.Any([]).tag(sync=True)           # SVG files to display (array of json objects containing "title" and "svg" tags)
    
    width     = traitlets.Unicode('80px').tag(sync=True)   # Width of the cards
    height    = traitlets.Unicode('100px').tag(sync=True)  # Height of the cards
    color     = traitlets.Unicode('white').tag(sync=True)  # Background color
    cols      = traitlets.Int(6).tag(sync=True)            # Horizontal column span [1,12] for each of the card
    ripple    = traitlets.Bool(False).tag(sync=True)       # Ripple flag (if True the click on the card is highlighted)
    on_click  = traitlets.Any(None).tag(sync=False)        # Name of a python function to call when one of the cards is clicked (it will receive as argument the index of the clicked card)


    @traitlets.default('template')
    def _template(self):
        return '''
<v-container fluid style="overflow: hidden;">
  <v-row dense>
    <v-col
      v-for="card,index in svgitems"
      :key="card.title"
      :cols="cols"
    >
        <v-card
            class="pa-0 ma-1"
            :width="width"
            :color="color"
            :height="height"
            hover
            :ripple="ripple"
            link
            elevation=0
            @click="clicked(index)"
        >
             <div style="display: flex; justify-content: center; align-items: center; overflow: hidden;" v-html="card.svg"/>
             <v-card-text class="pa-0 ma-0" style="font-size: 11; text-align: center;" v-html="card.title"/>
        </v-card>
    </v-col>
  </v-row>
</v-container>
'''

    def __init__(self, *args,
                 filepaths=[],
                 width='80px',
                 height='100px',
                 color='white',
                 cols=1,
                 ripple=False,
                 svgsize=80,
                 on_click=None,
                 **kwargs):
        
        self.svgitems = [{"svg": readSVGimg(x,svgsize), "title": os.path.basename(x).replace(".svg","").replace("_"," ")} for x in filepaths]
        self.width    = width
        self.height   = height
        self.color    = color
        self.cols     = cols
        self.ripple   = ripple
        self.on_click = on_click
        
        super().__init__(*args, **kwargs)
        
    
    # Manage "click" event
    def vue_clicked(self, data):
        if not self.on_click is None:
            self.on_click(data)


"""Menu widget opened on hover on a button."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import fontsettings
except:
    import settings
    import fontsettings

    
#####################################################################################################################################################
# Menu control
#####################################################################################################################################################
class menu:
    """
    Menu widget opened on hover on a button.
        
    Parameters
    ----------
    index : int
        Index of the option initially selected (from 0 to len(labels)-1)
    title : str
        Title string to display in the button
    labels : list of strings
        Strings to be displayed as options in the menu widget
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    onchange : function, optional
        Python function to call when the user selects one of the values in the list. The function will receive a single parameter, containing the index of the selected option in the range from 0 to len(labels)-1
    width : int, optional
        Width of the button in pixels (default is 150 pixels)
    highliteselection: bool, optional
        If True, the menu will show the selected option in bold (default is True)
            
    Example
    -------
    Creation and display of a menu widget to select among three options::
        
        from vois.vuetify import menu
        from ipywidgets import widgets
        from IPython.display import display
        
        output = widgets.Output()

        def onchange(value):
            with output:
                print(value)

        m = menu.menu(0, 'Hover to select',
                      ['Option 0', 'Option 1', 'Option 2'],
                      onchange=onchange, highliteselection=True)

        display(m.draw())
        display(output)

    .. figure:: figures/menu.png
       :scale: 100 %
       :alt: menu widget

       Example of a menu widget to select from three options.
   """

    # Initialization
    def __init__(self, index, title, labels, color=settings.color_first, onchange=None, width=150, highliteselection=True):
        
        self.index    = index
        self.title    = title
        self.labels   = labels
        self.color    = color
        self.onchange = onchange
        self.width    = width
        
        
        self.style_normal = 'font-family: %s; font-weight:400;' % (fontsettings.font_name)
        if highliteselection: self.style_selected = 'font-family: %s; font-weight:700;' % (fontsettings.font_name)
        else:                 self.style_selected = self.style_normal
        
        # Create the controls
        self.items = []
        i = 0
        for label in self.labels:
            item = v.ListItem(children=[label])
            if i == self.index: item.style_ = self.style_selected
            else:               item.style_ = self.style_normal
                
            self.items.append(item)
            item.on_event('click', self.__internal_onchange)
            i += 1

        self.menu = v.Menu(offset_y=True, open_on_hover=True, dense=True,
                           v_slots=[{
                                  'name': 'activator',
                                  'variable': 'menuData',
                                  'children': v.Btn(v_on='menuData.on', depressed=True, large=False, dense=True, class_='pa-0 ma-0', color=settings.color_first,
                                                    rounded=True, height=36, width=self.width,
                                                    style_='font-family: %s; text-transform: none' % (fontsettings.font_name), children=[self.title]),
                                   }],
                           children=[v.List(children=self.items)] )

    
    # Manage onchange event
    def __internal_onchange(self, widget=None, event=None, data=None):
        if self.index >= 0 and self.index < len(self.items):
            self.items[self.index].style_ = self.style_normal
        self.index = self.items.index(widget)
        self.items[self.index].style_ = self.style_selected
        if self.onchange:
            self.onchange(self.index)
    
    
    # Get the active option
    @property
    def value(self):
        """
        Get/Set the active option index.
        
        Returns
        --------
        index : int
            Index of the selected option (from 0 to len(labels)-1)

        Example
        -------
        Programmatically select one of the options and print the index of the selected option::
            
            m.value = 2
            print(m.value)
        
        """
        return self.index
   
    
    # Set the active option
    @value.setter
    def value(self, index):
        if index >= 0 and index < len(self.items):
            if self.index >= 0 and self.index < len(self.items):
                self.items[self.index].style_ = self.style_normal
            self.index = index
            self.items[self.index].style_ = self.style_selected
            if self.onchange:
                self.onchange(self.index)
                
                
    # Returns the vuetify object to display (the v.Menu itself)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Menu widget)"""
        return self.menu




"""Display of a Pandas DataFrame in a data-table widget."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
import traitlets
import pandas as pd
import json

# DataTable managing click on a row. See https://github.com/mariobuikhuizen/ipyvuetify/issues/163
class datatable(v.VuetifyTemplate):
    """
    Display of a Pandas DataFrame in a data-table widget.
        
    Parameters
    ----------
    data : Pandas DataFrame, optional
        Pandas DataFrame to be displayed (default is pd.DataFrame() empty DataFrame)
    height : str, optional
        Height of the data-table widget (default is '400px')
    on_click : function, optional
        Python function to call when the user clicks on one of the rows of the data-table. The function will receive a parameter of type dict containing all the column names as keys and the clicked row data as values
    color : str, optional
        Color to use for the display of alert and warning messages (for instance if no records are present in input DataFrame) (default is 'error')
    dark : bool, optional
        If True, the error and warning messages are displayed in white color, if False they are displayed in black (default is False)
    searchshow : bool, optional
        If True, on top of the table a search field will be displayed, allowing for search (default is False)
    search : str, optional
        Initial search string to be displayed in the search field (default is ''). If searchshow is False, this argument will not be used
    title : str, optional
        In case searchshow is True, a Title can be shown on top of the datatable (default is '')
    icon : str, optional
        In case searchshow is True, an icon can be shown on the right of the datatable title (default is '')
    icon_tooltip : str, optional
        Tooltip string to be used for the icon (default is '')
    icon_color : str, optional
        Color of the icon (default is 'grey')
    icon_disabled : bool, optional
        If True, the icon will be disabled (default is False)
    font_size : str, optional
        Font size to use for the rows and headers of the datatable (default is '14px')
    font_size_title : str, optional
        Font size to use for the title of the datatable (default is '16px')
    on_icon : function, optional
        Python function to call when the user clicks on the icon on the top of the data-table. The function will receive no parameters
    unsortable_columns : list of str, optional
        List of names of columns that must not be sortable in the datatable (please note that the DataFrame column names will be displayed in capital letters in the datatable: use in this list the original column names of the DataFrame and not the capitalized names)

    Example
    -------
    Creation of a Pandas DataFrame from the 'Our World In Data' dataset on Covid-19 daily data and display of last 100 days for Italy::
        
        from vois.vuetify import datatable
        import pandas as pd
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        df = pd.read_csv('https://raw.githubusercontent.com/owid/covid-19-data/master/public/data/owid-covid-data.csv')

        def on_click(data):
            output.clear_output()
            with output:
                print(data)

        df = df[df['location']=='Italy']
        d = datatable.datatable(data=df.tail(100), height='500px', on_click=on_click)

        display(d)
        display(output)

    .. figure:: figures/datatable.png
       :scale: 100 %
       :alt: datatable widget

       Last 100 days of Covid-19 data on Italy displayed in a datatable widget
    """
    
    headers = traitlets.List([]).tag(sync=True, allow_null=True)
    items = traitlets.List([]).tag(sync=True, allow_null=True)
    index_col = traitlets.Unicode('').tag(sync=True)
    height = traitlets.Unicode('400px').tag(sync=True)
    color = traitlets.Unicode('error').tag(sync=True)  # Color for error message in case of empty input DF
    dark = traitlets.Bool(False).tag(sync=True)
    
    searchshow = traitlets.Bool(False).tag(sync=True)
    title      = traitlets.Unicode('').tag(sync=True)
    search     = traitlets.Unicode('').tag(sync=True)
    
    icon          = traitlets.Unicode('').tag(sync=True)
    icon_tooltip  = traitlets.Unicode('').tag(sync=True)
    icon_color    = traitlets.Unicode('grey').tag(sync=True)
    icon_disabled = traitlets.Bool(False).tag(sync=True)
    
    font_size       = traitlets.Unicode('14px').tag(sync=True)
    font_size_title = traitlets.Unicode('16px').tag(sync=True)
    
    @traitlets.default('template')
    def _template(self):
        
        if self.searchshow:
            
            stricon = ''
            if len(self.icon) > 0:
                stricon = '''
    <v-tooltip bottom>
      <template v-slot:activator="{ on, attrs }">
        <v-btn icon @click="onicon()" :disabled="icon_disabled" color="%s" v-on="on"><v-icon small>%s</v-icon></v-btn>
      </template>
      <span>%s</span>
    </v-tooltip>
'''% (self.icon_color, self.icon, self.icon_tooltip)

                
            cardtitle = '''
            <v-card-title class="pa-0 ma-0 ml-2 mb-1" style="font-size: %s;">
              %s
              %s
              <v-spacer></v-spacer>
              <v-text-field class="pa-0 ma-0 mb-3" v-model="search" :color="color" append-icon="mdi-magnify" label="Search" single-line hide-details></v-text-field>
            </v-card-title>
            ''' % (self.font_size_title, self.title, stricon)
        else:
            cardtitle = ''
        
        return '''
        <template>
        <v-card>
            %s
            <v-data-table
                dense
                hide-default-footer
                fixed-header
                :search="search"
                :height="height"
                :headers="headers"
                :items="items"
                :item-key="index_col"
                :footer-props="{'items-per-page-options': [10000000]}">
                <template v-slot:no-data> 
                  <v-alert :value="true" :color="color" :dark="dark" icon="mdi-alert">
                    No records to display
                  </v-alert>
                </template>
                <template v-slot:no-results>
                    <v-alert :value="true" :color="color" :dark="dark" icon="mdi-alert" width="80vw">
                      Your search for "{{ search }}" found no results
                    </v-alert>
                </template>

                <template v-slot:item="row">
                  <tr>
                    <td v-for="value in Object.values(row.item)"
                        @click="cell_click(row.item)"
                    >
                      {{ value }}
                    </td>
                  </tr>
                </template>

            </v-data-table>
          </v-card>
        </template>

<style>
.v-data-table > .v-data-table__wrapper > table > tbody > tr > th,
.v-data-table > .v-data-table__wrapper > table > thead > tr > th,
.v-data-table > .v-data-table__wrapper > table > tfoot > tr > th,
.v-data-table > .v-data-table__wrapper > table > tbody > tr > td {
   font-size: %s !important;
}
</style>
''' % (cardtitle, self.font_size)
    
    on_click = None  # This must receive a function which will be called inside `self.vue_cell_click()`

    # Click on the "icon" button
    def vue_onicon(self, data):
        if not self.on_icon is None:
            self.on_icon()
    
    # Click on one row of the datatable
    def vue_cell_click(self, data):
        if not self.on_click is None:
            self.on_click(data)
    
    # Initializing
    def __init__(self, *args,
                 data=pd.DataFrame(),
                 height='400px',
                 title='',
                 dictwidth={},
                 searchshow=False,
                 search='',
                 on_click=None, 
                 on_icon=None,
                 icon='',
                 icon_tooltip='',
                 icon_color='grey',
                 icon_disabled=False,
                 font_size='14px',
                 font_size_title='16px',
                 unsortable_columns=[],
                 **kwargs):
        
        self.title = title
        self.searchshow = searchshow
        self.search = search
        
        data = data.reset_index()
        self.index_col = data.columns[0]
       
        #headers = [{"text": col, "value": col } for col in data.columns]
        headers = []
        for col in data.columns:
            
            sortable = True
            if col in unsortable_columns: 
                sortable = False
                
            h = {"text": str(col).upper(), "value": col, "sortable": sortable}
            
            if col in dictwidth:
                h["width"] = dictwidth[col]
            else:
                h["width"] = "20px"
            headers.append(h)
        #print(headers)
        
        self.headers = headers
        self.height = height
        self.on_click = on_click
        self.on_icon = on_icon
        self.icon = icon
        self.icon_tooltip = icon_tooltip
        self.icon_color = icon_color
        self.icon_disabled = icon_disabled
        self.font_size = font_size
        self.font_size_title = font_size_title
        self.items = json.loads(data.to_json(orient='records'))

        super().__init__(*args, **kwargs)



"""Footer bar to be displayed at the bottom of a Voilà dashboard."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import tooltip
except:
    import settings
    import tooltip


#####################################################################################################################################################
# Display a footer panel with some buttons
#####################################################################################################################################################
class footer():
    """
    Footer bar to be displayed at the bottom of a Voilà dashboard.
        
    Parameters
    ----------
    text : str, optional
        Text message to display in the bottom line of the footer bar (default is '')
    copyright : bool, optional
        Flag to display a copyright symbol in the bottom line of the footer bar (default is False)
    buttons : list of strings, optional
        List of strings to use for adding a row of buttons (default is []])
    color : str, optional
        Background color of the title bar (default is the main color defined in the settings.py module)
    dark : bool, optional
        Flag that controls the color of the text in foreground: if True, the text will be displayed in white, elsewhere in black (default is the value of settings.dark_mode)
    height : int or float or str optional
        Height of the footer bar. If an integer or a float is passed, the height is intended in pixels units, otherwise a string containing the units must be passed (example: '4vh'). Default is 68 for 68 pixels
    marginy : int, optional
        Vertical displace of the first row of the bar from the top (default is 2)
    onclick : function, optional
        Python function to call when the user clicks on one of the buttons. The function will receive a parameter of type str containing the text of the clicked button
    output : ipywidgets.Output, optional
        Output widget on which the footer bar has to be displayed
    minipanelicons : list of strings, optional
        Names of optional icons to display in the minipanel at the left side of the footer bar
    minipaneltooltips : list of strings, optional
        Tooltips of optional icons to display in the minipanel at the left side of the footer bar
    minipanelopen : bool, optional
        If True the minipanel is initially displayed already opened (default is False)
    minipanellarge : bool, optional
        If True the icons in theminipanel are displayed with greater dimension (default is True)
    minipanelbuttontooltip: str, optional
        Text of the tooltip to show when hover on the minipanel open icon (default is 'Additional functions')
    minipanelbuttoncolor : str, optional
        Color of the icon that opens/closes the minipanel (default is the textcolor_notdark defined in the settings.py module)
    minipaneliconscolor : str, optional
        Color of the icons inside the minipanel (default is the textcolor_notdark defined in the settings.py module)
    footercredits : str, optional
        Text for footer credits button (default is '')
    footercreditstooltip : str, optional
        Tooltip for the footer credits button (default is '')
    onclickcredits : function, optional
        Python function to call when the user clicks on the credits button. The function will be called with 0 parameters
    onclickminipanel : function, optional
        Python function to call when the user clicks on one of the icons of the minipanel. The function will receive a parameter of type int containing the index of the clicked icon in the range from 0 to len(minipanelicons)-1
            
    Example
    -------
    Creation and display of a footer bar::
        
        from vois.vuetify import footer
        from ipywidgets import widgets
        from IPython.display import display
        from datetime import datetime

        output = widgets.Output()
        display(output)

        def onclick(arg):
            with output:
                print(arg)

        def onclickminipanel(index):
            with output:
                print(index)

        def onclickcredits():
            with output:
                print('CREDITS')

        f = footer.footer(text='%d - Joint Research Centre'%(datetime.now().year), color='lightgrey',
                          minipanelicons=['fa-truck', 'mdi-heart', 'mdi-magnify'],
                          minipaneltooltips=['Function 1', 'Function 2', 'Function 3'], 
                          minipanellarge=True, minipanelopen=True,
                          onclickminipanel=onclickminipanel,
                          footercredits='Data credits',
                          footercreditstooltip='Eurostat - European Commission',
                          onclickcredits=onclickcredits,
                          buttons=['Home', 'About Us', 'Services', 'Contact Us'],
                          height=68, marginy=2,
                          onclick=onclick,
                          output=output)
                  
    .. figure:: figures/footer.png
       :scale: 100 %
       :alt: footer widget

       Footer bar with buttons and minipanel.
       
    Note
    ----
    The footer component is used inside the :py:class:`app.app` class. If the main interface of a dashboard is created using an instance of the :py:class:`app.app` class, the parameters for customising the footer bar can be passed in the construction of the **app** instance.
    
   """
        
    def __init__(self, text='', copyright=False, buttons=[], dark=settings.dark_mode, height=68, marginy=2, color=settings.color_first, 
                 minipanelicons=[], minipaneltooltips=[], minipanelopen=False,
                 minipanellarge=True, minipanelbuttontooltip='Additional functions',
                 minipanelbuttoncolor=settings.color_first, minipaneliconscolor=settings.textcolor_notdark,
                 footercredits='',        # Text for footer credits button
                 footercreditstooltip='', # Tooltip for the footer credits button
                 onclickcredits=None,     # Function with 0 parameters
                 onclickminipanel=None,   # Function with 1 string argument integer containing the index of the icon
                 onclick=None,            # Function with 1 string argument integer containing the text of the button
                 output=None):
        
        self.onclick = onclick
        self.onclickminipanel = onclickminipanel
        self.onclickcredits = onclickcredits
        
        if dark:
            textcol = settings.textcolor_dark
            if minipaneliconscolor == settings.textcolor_notdark:
                minipaneliconscolor = settings.textcolor_dark
        else:
            textcol = settings.textcolor_notdark
            
        members = [text]
        if copyright:
            icon = v.Icon(class_="pa-0 ma-0 mr-2", small=True, color=textcol, children=['mdi-copyright'])
            members = [icon,text]

        if len(minipanelicons) > 0:
            if len(buttons) == 0:
                c = v.Col(color=color, class_="pa-0 ma-0 mt-%d text-center %s--text" % (marginy+4,textcol), cols="12", children=members)
            else:
                c = v.Col(color=color, class_="pa-0 ma-0 text-center %s--text" % (textcol), cols="12", children=members)
        else:
            c = v.Col(color=color, class_="pa-0 ma-0 mt-%d text-center %s--text" % (marginy,textcol), cols="12", children=members)
    
        self.buttons_text = buttons
        self.buttons = [v.Btn(color=textcol, text=True, rounded=settings.button_rounded, class_='my-2', children=[x]) for x in self.buttons_text]
        for b in self.buttons:
            b.on_event('click', self.__internal_onclick)
            
        r = v.Row(class_='pa-0 ma-0 mt-n2 mb-n3', justify="center", no_gutters=True, children=self.buttons)

        if self.onclickcredits is None or len(footercredits) == 0:
            self.credits = v.Btn(text=True, color=textcol, dark=dark, children=[footercredits], rounded=settings.button_rounded, disabled=True, style_='text-transform: none; cursor: initial;')
        else:
            self.credits = v.Btn(text=True, color=textcol, dark=dark, children=[footercredits], rounded=settings.button_rounded, style_='text-transform: none;')
            self.credits.on_event('click', self.__internal_onclickCredits)
        if len(footercreditstooltip) > 0:
            self.credits = tooltip.tooltip(footercreditstooltip,self.credits)

        # Minipanel
        largeicon = minipanellarge
        panelwidth = 40
        iconclass = "pa-0 ma-0 mr-1"
        panelw  = (panelwidth)*(len(minipanelicons))
        panelh = panelwidth-panelwidth/10
        if largeicon:
            panelwidth = 50
            iconclass = "pa-0 ma-0 mr-2"
            panelw  = (panelwidth)*(len(minipanelicons)) - 16
            panelh = panelwidth-12

        self.minipanel_btn = [v.Btn(icon=True, class_=iconclass, children=[v.Icon(large=largeicon, color=minipaneliconscolor, children=[x])]) for x in minipanelicons]
        for b in self.minipanel_btn:
            b.on_event('click', self.__internal_onclick_minipanel)
        minipanel_tb = [tooltip.tooltip(x,y) for x,y in zip(minipaneltooltips,self.minipanel_btn)]
        
        if len(self.minipanel_btn) > 0:
            if isinstance(height, int) or isinstance(height, float):
                hhh = '%fpx' % height
            else:
                hhh = str(height)
            y = 'calc(calc(%s - %dpx) / 2)' % (hhh,panelh)
            navDrawer = v.NavigationDrawer(v_model=minipanelopen, children=[v.Row(justify="center", no_gutters=True, children=minipanel_tb)],
                                           width="%dpx"%panelw, height="%dpx"%panelh,
                                           floating=True, fixed=True, absolute=True, style_="left:40px; top:%s; z-index:999;"%y)
            navDrawer.mini_variant = True
            navDrawer.expand_on_hover = False
            navDrawer.width = panelw
            navDrawer.mini_variant_width = panelw
            
            toolBarButton = v.Btn(icon=True, class_='pa-0 ma-0 ml-1', children=[v.Icon(children=['mdi-dots-vertical'])])
            def on_click(widget, event, data):
                navDrawer.v_model = not navDrawer.v_model
                if navDrawer.v_model: toolBarButton.color = minipanelbuttoncolor
                else:                 toolBarButton.color = textcol
                    
            if navDrawer.v_model: toolBarButton.color = minipanelbuttoncolor
            else:                 toolBarButton.color = textcol
                
            toolBarButton.on_event('click', on_click)
            tbt = tooltip.tooltip(minipanelbuttontooltip,toolBarButton)
            c1 = v.Col(cols="1", align_self="center", children=[tbt, navDrawer])
            c2 = v.Col(cols="10",children=[r,c])
            c3 = v.Col(cols="1", align_self="right", children=[self.credits])
            footercontent = [v.Row(justify="start", no_gutters=True, children=[c1,c2,c3])]
        else:
            if len(self.buttons) > 0:
                c2 = v.Col(cols="11",children=[r,c])
            else:
                c2 = v.Col(cols="11",children=[c])
            c3 = v.Col(cols="1", align_self="right", children=[self.credits])
            footercontent = [v.Row(justify="start", no_gutters=True, children=[c2,c3])]
        
        hhh = '68px'
        if not height is None:
            if isinstance(height, int) or isinstance(height, float):
                hhh = '%fpx' % height
            else:
                hhh = height
                
        self.f = v.Footer(color=color, padless=True, children=footercontent, class_='pa-0 ma-0', style_='height: %s; overflow: hidden;'%hhh)
        
        # Display of the footer
        if not output is None:
            with output:
                display(self.f)

    # Returns the vuetify object to display (the v.Footer)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Html which has a v.Footer as its only child)"""
        return v.Html(tag='div',children=[self.f])
                
    # Manage click on the buttons
    def __internal_onclick(self,widget,event,data):
        i = self.buttons.index(widget)
        if not self.onclick is None:
            self.onclick(self.buttons_text[i])
            
            
    # Manage click on the minipanel icons
    def __internal_onclick_minipanel(self,widget,event,data):
        i = self.minipanel_btn.index(widget)
        if not self.onclickminipanel is None:
            self.onclickminipanel(i)

            
    # Manage click on the title credits button
    def __internal_onclickCredits(self, widget, event, data):
        if not self.onclickcredits is None:
            self.onclickcredits()
            
    


"""Label widget to display a text with an optional icon."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import tooltip
    from . import fontsettings
except:
    import settings
    import tooltip
    import fontsettings


#####################################################################################################################################################
# Label class that can display also an icon
#####################################################################################################################################################
class label:
    """
    Label widget to display a text with an optional icon.
        
    Parameters
    ----------
    text : str
        Test string to be displayed on the label widget
    onclick : function, optional
        Python function to call when the user clicks on the label. The function will receive as parameter the value of the argument (default is None)
    argument : any, optional
        Argument to be passed to the onclick function when user click on the label (default is None)
    disabled : bool, optional
        Flag to show the label as disabled (default is False)
    textweight : int, optional
        Weight of the text to be shown in the label (default is 350, Bold is any value greater or equal to 500)
    height : int, optional
        Height of the label widget in pixels (default is 20)
    margins : int, optional
        Dimension of the margins on all directions (default is 0)
    margintop : int, optional
        Dimension of the margin on top of the label (default is None)
    icon: str, optional
        Name of the icon to display aside the text of the label (default is None)
    iconlarge : bool, optional
        Flag that sets the large version of the icon (default is False)
    iconsmall : bool, optional
        Flag that sets the small version of the icon (default is False)
    iconleft : bool, optional
        Flag that sets the position of the icon  to the left of the text of the label (default is False)
    iconcolor : str, optional
        Color of the icon (default is 'black')
    tooltip : str, optional
        Tooltip string to display when the user hovers on the label (default is None)
    textcolor : str, optional
        Color used for the label text
    backcolor : str, optional
        Color used for the background of the label
    dark : bool, optional
        Flag to invert the text and backcolor (default is the value of settings.dark_mode)
            

    Note
    ----
    All the icons from https://materialdesignicons.com/ site can be used, just by prepending 'mdi-' to their name.
    
    All the free icons from https://fontawesome.com/ site can be used, just by prepending 'fa-' to their name.


    Example
    -------
    Creation and display of a label widget containing an icon::
        
        from vois.vuetify import label

        lab = label.label('Test label', textweight=300, margins=2,
                          icon='mdi-car-light-high', iconcolor='red',
                          iconlarge=True, height=22)

        display(lab.draw())
    
    
    .. figure:: figures/label.png
       :scale: 100 %
       :alt: label widget

       Example of a label widget with text and an icon.
   """

   
    # Initialization
    def __init__(self, text, onclick=None, argument=None, disabled=False, textweight=350, height=20, margins=0, margintop=None, 
                 icon=None, iconlarge=False, iconsmall=False, iconleft=False, iconcolor='black', tooltip=None,
                 textcolor=None, backcolor=None, dark=settings.dark_mode):
        
        self.labeltext = text
        self.onclick   = onclick
        self.argument  = argument
        
        self.disabled   = disabled
        self.textweight = textweight
        self.height     = height
        self.margins    = margins
        self.margintop  = margintop
    
        self.icon       = icon
        self.iconleft   = iconleft
        self.iconlarge  = iconlarge
        self.iconsmall  = iconsmall
        self.iconcolor  = iconcolor

        self.tooltip    = tooltip
        
        self.textcolor  = textcolor
        self.backcolor  = backcolor
        self.dark       = dark
        
        if not self.dark is None:
            if self.dark:
                if self.backcolor is None: self.backcolor='#111111'
                if self.textcolor is None: self.textcolor='white'

        self.__createLabel()
       
        
    # Create the label
    def __createLabel(self):
        if self.icon is None:
            childs = [self.labeltext]
        else:
            childs = [self.labeltext, v.Icon(left=self.iconleft, large=self.iconlarge, small=self.iconsmall, color=self.iconcolor, children=[self.icon])]
            
        strstyle = 'font-family: %s; font-size: 10; font-weight: %d; text-transform: none;' % (fontsettings.font_name, self.textweight)
        if not self.textcolor is None:
            strstyle += 'color: %s;' % self.textcolor
        if not self.backcolor is None:
            strstyle += 'background-color: %s;' % self.backcolor
        
        self.item = v.Card(disabled=self.disabled, elevation=0, height=self.height, depressed=True, children=childs, style_=strstyle)
        if not self.tooltip is None:
            self.item = tooltip.tooltip(self.tooltip, self.item)
            
        
        # If requested onclick management
        if not self.onclick is None:
            self.item.on_event('click', self.__internal_onclick)
            
        if not self.margintop is None:
            self.container = v.Container(class_="pa-0 ma-%s mt-%s" % (str(self.margins), str(self.margintop)), children=[ self.item ])
        else:
            self.container = v.Container(class_="pa-0 ma-%s" % (str(self.margins)), children=[ self.item ])
        
        
    # Manage click event
    def __internal_onclick(self, widget=None, event=None, data=None):
        if self.onclick:
            if not self.argument is None:
                self.onclick(self.argument)
            else:
                self.onclick()
    
    # Returns the vuetify object to display (the v.Container)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Html which has a v.Container as its only child)"""
        return v.Html(tag='div',children=[self.container])
    
    
    # Get the label text
    @property
    def text(self):
        """
        Get/Set the label text.
        
        Returns
        --------
        text : str
            Text currently shown in the label

        Example
        -------
        Programmatically set the label text (needs a re-display to be visible!)::
            
            lab.text = 'New text for the label'
            display(lab.draw())
            print(lab.text)
        
        """
        return self.labeltext
   
    
    # Set the label text
    @text.setter
    def text(self, newtext):
        self.labeltext = newtext
        self.__createLabel()
            



"""Class to display and query raster datasets from BDAP (JRC Big Data Analytics Platform) using the interapro library MapnikRaster class."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

from jeodpp import inter

# Python imports
from io import StringIO
import sys



#####################################################################################################################################################
# Class raster
#####################################################################################################################################################
class raster:
    
    # Initialization
    def __init__(self,
                 filepath,
                 band=1,
                 epsg=4326,
                 nodata=999999.0):

        self.filepath = filepath
        self.band     = band
        self.epsg     = epsg
        self.nodata   = nodata

        self.r = inter.MapnikRaster()
        self.r.file(self.filepath, band=self.band, epsg=self.epsg, nodata=self.nodata)
        
    
    # Representation
    def __repr__(self):
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        self.r._print()
        sys.stdout = old_stdout
        return mystdout.getvalue()
        
        
    # Create a symbolizer: see https://github.com/mapnik/mapnik/wiki/RasterSymbolizer
    def symbolizer(self,
                   scaling="near",
                   opacity=1.0,
                   composition=""):
        self.r.rasterSymbolizer(scaling, opacity, composition)
        
        
    # Create a colorizer: see https://github.com/mapnik/mapnik/wiki/RasterColorizer
    def colorizer(self,
                  default_mode="linear",
                  default_color="transparent",
                  epsilon=1.5e-07):
        
        self.r.rasterColorizerInit(default_mode,default_color,epsilon)
        

    # Add a colorizer step: see https://github.com/mapnik/mapnik/wiki/RasterColorizer#example-xml
    def color(self,
              value,            # Numerical value
              color="red",      # name of color or "#rrggbb"
              mode=""):         # "discrete", "linear" or "exact"
        
        self.r.rasterColorizerStep(value, color, mode)
        
        
    # Return the ImageProcess instance 
    @property
    def ip(self):
        return self.r
    
    # Return the XML string containing the Mapnik map
    def xml(self):
        return self.r.getXML()
    
    
    # Save the layer in Redis and returns the procid
    def toLayer(self):
        self.r.setXML()
        return self.r.toLayer()
    
    # Return the JSON string defining the raster layer
    def toJson(self):
        self.r.setXML()
        return self.r.toJson()    


"""SVG bubbles chart from a pandas DataFrame."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import pandas as pd
from datetime import datetime
from dateutil.rrule import rrule, DAILY, WEEKLY, MONTHLY

from ipywidgets import HTML, widgets, Layout
from ipyevents import Event
from IPython.display import display

import plotly.express as px

try:
    from . import colors
    from .vuetify import fontsettings
except:
    import colors
    from vuetify import fontsettings


# Returns a list of strings (splitted at ' ')
def splitstring(s, maxlen):
    nrows = int( (len(s)+maxlen-1) / maxlen )
    s = s.replace('_',' ')
    s = s.replace('-',' ')
    words = s.split(' ')

    rows = ['']
    pos = 0
    for www in words:
        if len(rows[pos]) <= maxlen - len(www):
            if len(rows[pos]) == 0:  rows[pos] = www
            else:                    rows[pos] += ' ' + www
        else:
            rows.append(www)
            pos += 1
            
    return rows

                            
# Vertical text in an SVG where X coordinate are in percentage!!!
def verticalText(text, x, y, size, color, fontfamily, weight):
    return '''
<text x="%fvw" y="%fvh" font-size="%fvh" fill="%s" font-weight="%d" style="font-family: %s; pointer-events: none; transform: rotate(90deg); transform-origin: %fvw %fvh; transform-box: view-box; overflow: visible;">%s</text>
''' % (x, y, size, color, weight, fontfamily, x, y, text)


###########################################################################################################################################################################
# Bubbles chart visualization of 3 discrete variables + a size
###########################################################################################################################################################################
class svgBubblesChart:
    """
    Creation of a bubbles chart given an input DataFrame. It is a convenient chart for representing a numerical value that depends on 3 discrete variables. It displays a bi-dimensional grid where the unique values of the x column are displayed on the X axis, while the unique values of the y column are displayed on the Y axis. Inside each cell of the grid, a group of bubbles is displayed, one for each distinct value of the color column, while the size of the circles is proportional to the numerical value read from the size column. See the below example on a mushrooms dataset (taken from https://www.kaggle.com/datasets/uciml/mushroom-classification). The SVG chart has the x coordinates expressed in vw coordinates and the y coordinates expressed in vh coordinates. Cliks on the legend is managed so that individual color categories can be excluded from the chart.
    
    Parameters
    ----------
    df : pandas DataFrame
        Input DataFrame
    xcolumn : str
        Name of the DataFrame column containing the values to be displayed on the X axis
    ycolumn : str
        Name of the DataFrame column containing the values to be displayed on the X axis
    sizecolumn : str
        Name of the DataFrame column containing the numerical values to be used to size the bubbles
    colorcolumn : str
        Name of the DataFrame column containing the values to be used to give a color to the bubbles and to display the color legend
    width : float, optional
        Width of the chart in vw units (default is 100.0)
    height : float, optional
        Height of the chart in vh units (default is 50.0)
    xstart : float, optional
        X coordinate on vw units where the grid starts (default is 6.0). This value can be used to leave more or less space to the Y axis strings
    strokewidth : int, optional
        Width in pixels of the stroke to use for the bubbles (default is 1)
    strokecolor : str, optional
        Color of the stroke to use for the bubbles (default is '#ffffff')
    backcolor : str, optional
        Background color of the grid (default is '#aaaaaa')
    backlinecolor : str, optional
        Color of the lines defining the cells of the grid (default id '#888888')
    bubblecolors : list of colors, optional
        List of colors to use for the bubble. Each color is assigned to one of the unique values of the color column of the input DataFrame (default is px.colors.qualitative.Dark2)
    textcolor : str, optional
        Color to use for the texts, chart title, axis titles, axis values, etc. (default is 'black')
    textweight : int, optional
        Weight of the text (default is 400). The chart title, the axis titles and the legend title will be displayed with weight equal to textweight+100
    fontsize : float, optional
        Size of the standard font to use for values displayed in the X and Y axis in vh coordinates (default is 1.1vh). The chart title, the axis titles and the legend title will be displayed with sizes proportional to the fontsize parameter (up to two times for the chart title)
    xtextangle : float, optional
        Angle to use for the rotation of values displayed on the X axis (default is 0.0)
    title : str, optional
        Title of the chart (default is '')
    mode : str, optional
        Mode to use for the placement of bubbles inside the cells of the grid. Possible values are 'spread' (circles are horizontally displaced), 'concentric' (all circles have their centre in the  center of the cell) or 'tangent' (circles are all tangent on the center-bottom point of the cell). Default is 'spread'.
    legendrows : int, optional
         Number of rows to use for the legend (default is 2)
    legenditemwidth : int, optional
        Width in percent of the total width of each item of the legend (default is 10)
            
    Returns
    -------
        an instance of widgets.Output with the svg chart displayed in it

    Example
    -------
    Creation of a SVG bubble chart chart to display some of the columns of the mushroom dataset (see https://www.kaggle.com/datasets/uciml/mushroom-classification)::

        from IPython.display import display
        import pandas as pd
        import plotly.express as px
        colorlist = px.colors.qualitative.Dark2
        from vois import svgBubblesChart

        df = pd.read_csv('https://jeodpp.jrc.ec.europa.eu/services/shared/csv/mushrooms.csv')

        xcolumn     = 'cap-shape'
        ycolumn     = 'cap-surface'
        colorcolumn = 'habitat'
        sizecolumn  = 'count'
        dfgrouped = df.groupby([xcolumn, ycolumn, colorcolumn]).size().reset_index(name=sizecolumn)

        dfgrouped.columns = ['Shape', 'Cap surface', 'Mushroom habitat:', 'count']

        b = svgBubblesChart.svgBubblesChart(dfgrouped,
                                            height=50.0,
                                            xcolumn=dfgrouped.columns[0], 
                                            ycolumn=dfgrouped.columns[1],
                                            colorcolumn=dfgrouped.columns[2],
                                            sizecolumn=dfgrouped.columns[3], 
                                            strokewidth=1,
                                            strokecolor='#ffffff', 
                                            backcolor='#f0f0f0',
                                            backlinecolor='#000000',
                                            bubblecolors=colorlist,
                                            fontsize=1.1,
                                            title='Mushrooms analysis',
                                            mode='spread')
        display(b.draw())
        display(HTML(b.getlegendsvg()))
        
    .. figure:: figures/bubblechart.png
       :scale: 100 %
       :alt: svgBubblesChart example

       Example of an interactive bubbles chart in SVG
        
    """

    # Initialization: width in vw units, height in vhunits
    def __init__(self, df, xcolumn, ycolumn, sizecolumn, colorcolumn, width=100.0, height=50.0, xstart=6.0, 
                 strokewidth=1, strokecolor='#ffffff', backcolor='#aaaaaa', backlinecolor='#888888', 
                 bubblecolors=px.colors.qualitative.Dark2,
                 textcolor='black', textweight=400, fontsize=1.1, xtextangle=0.0, title='',
                 mode='spread',            # 'spread' or 'concentric' or 'tangent'
                 legendrows=2,             # Number of rows to use for the legend
                 legenditemwidth=10,       # Width in percent of each item of the legend
                ):
        
        self.df                = df
        self.xcolumn           = xcolumn
        self.ycolumn           = ycolumn
        self.sizecolumn        = sizecolumn
        self.colorcolumn       = colorcolumn
        if not self.colorcolumn is None and len(self.colorcolumn) <= 0: self.colorcolumn = None
        self.width             = width
        self.height            = height
        self.strokewidth       = strokewidth
        self.strokecolor       = strokecolor
        self.backcolor         = backcolor
        self.backlinecolor     = backlinecolor
        self.colorlist         = [colors.rgb2hex(colors.string2rgb(x)) for x in bubblecolors]   # Convert all colors to '#rrggbb' format !!!
        self.textcolor         = textcolor
        self.textweight        = textweight
        self.fontsize          = fontsize
        self.xtextangle        = xtextangle
        self.title             = title
        self.mode              = mode
        self.legendrows        = legendrows
        self.legenditemwidth   = legenditemwidth
        

        self.nrows   = self.df.shape[0]
        self.xvalues = sorted(list(self.df[self.xcolumn].unique()))
        self.yvalues = sorted(list(self.df[self.ycolumn].unique()))
        self.nx      = len(self.xvalues)
        self.ny      = len(self.yvalues)

        self.maxlenx = 20    # Max number of characters in xvalues: otherwise --> split
        self.maxleny = 12    # Max number of characters in yvalues: otherwise --> split
        
        if len(self.xvalues) > 15: self.maxlenx = 12
        nrows = 1
        for xvalue in self.xvalues:
            if len(xvalue) > self.maxlenx:
                rows = splitstring(xvalue,self.maxlenx)
                if len(rows) > nrows: nrows = len(rows)
        
        self.xvaluesindex = {}
        for x in self.xvalues: self.xvaluesindex[x] = self.xvalues.index(x)
        
        self.yvaluesindex = {}
        for y in self.yvalues: self.yvaluesindex[y] = self.yvalues.index(y)
            

        self.colorvalues = []
        self.ncolors = 0
        self.display = []
        if not self.colorcolumn is None:
            self.colorvalues = sorted(list(self.df[self.colorcolumn].unique()))
            self.ncolors     = len(self.colorvalues)
            self.display     = [True] * self.ncolors  # Flag to display or not a series

            
        self.legenditemheight  = 3.5
        self.titlefontsize     = self.fontsize*2
        #self.totalheight       = self.fontsize + self.height + (min(self.legendrows,self.ncolors) + 1)*self.legenditemheight + 1.4
            
        self.minvalue = 0.0
        if self.nrows > 0:
            self.maxvalue = float(self.df[self.sizecolumn].max())
        else:
            self.maxvalue = 0.0
            
            
        self.ystart   = 0.08 + self.fontsize*3                # In vh
        self.yend     = self.height - self.fontsize*2- (min(self.legendrows,self.ncolors) + 1)*self.legenditemheight
        
        self.xstart = xstart     # In vw
        self.xend   = self.width
        if self.nx > 0:
            self.xspace = (self.xend - self.xstart) / float(self.nx)
        else:
            self.xspace = 0.0
        
            
        if self.ny > 0:
            self.yspace = (self.yend - self.ystart) / float(self.ny)
        else:
            self.yspace = 0.0

        #self.rmax = min(self.xspace, self.yspace)
        self.rmax = self.yspace/2.0 - 0.1
        self.rmin = self.rmax/20.0
        
        # Create Output and Event
        self.out = widgets.Output(layout=Layout(width='calc(%fvw + 10px)'%self.width, height='calc(%fvh + 20px)'%self.height, margin='0px 0px 0px 0px'))  #, border='1px dashed green'))
        self.event = Event(source=self.out, watched_events=['click']) #, 'dblclick'])
        self.event.on_dom_event(self.handle_event)
        
        self.debug = widgets.Output()
        

    # Management of 'click' event on the SVG Output
    def handle_event(self, event):
        x = event['relativeX']
        y = event['relativeY']
        w = event['boundingRectWidth']
        h = event['boundingRectHeight']
        xp = self.width  * (x / w)
        yp = self.height * (y / h)
            
        # Click on a legend item
        if yp >= self.yend:

            for i in range(len(self.legendpos)):
                xitem = self.legendpos[i][0]
                yitem = self.legendpos[i][1]-1.5

                if yp >= yitem and yp <= yitem+self.legenditemheight and xp >= xitem and xp <= (xitem+1.5):
                    item = i
                    if item >= 0 and item < len(self.display):
                        if event['type'] == 'click':
                            self.display[item] = not self.display[item]
                        self.draw()
                    break
        
            
    # Return radius from a size value
    def getradius(self, size):
        if (self.maxvalue-self.minvalue) < 0.001: return self.rmax
        return self.rmin + (self.rmax - self.rmin) * (size - self.minvalue)/(self.maxvalue-self.minvalue)
    
    
    
    # Returns an SVG for the size legend
    def getlegendsvg(self):
        """
        Returns a string containing the SVG legend for the sizes of the circles.
        """
        titlefontsize = self.fontsize*1.4
        y = self.rmax + 2*titlefontsize
        fontsize = self.rmax/4.0
        
        sidetext = False
        if fontsize < 1.0:
            sidetext = True
            fontsize = 1.0
            w = self.yspace*1.3
        else:
            w = self.yspace
            
        h = w * 2.2
        svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0" y="0" width="%fvw" height="%fvh">' % (w,h)
        
        x = w/2.0
        svg += '<text text-anchor="middle" x="%fvw"  y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">Size legend:</text>' % (x-0.01, titlefontsize+0.6, titlefontsize,
                                                                                                                                                           self.textcolor, fontsettings.font_name, self.textweight+100)

            
        if sidetext:
            svg += '<text text-anchor="start" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%d</text>' % (x+self.rmax-0.3, y+fontsize/2-0.2, fontsize,
                                                                                                                                                    self.textcolor, fontsettings.font_name, self.textweight, int(self.maxvalue))
        else:
            svg += '<text text-anchor="middle" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%d</text>' % (x-0.01,y+fontsize/2-0.2, fontsize, self.textcolor, fontsettings.font_name, self.textweight, int(self.maxvalue))
        svg += '<circle cx="%fvw" cy="%fvh" r="%fvh" style="fill:none; stroke:black; stroke-width:1.5; opacity:1.0;"></circle>' % (x,y, self.rmax)
        
        if sidetext:
            y += self.rmax * 1.8
            svg += '<text text-anchor="start" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%d</text>' % (x+self.rmax-0.3,y+fontsize/2-0.2, fontsize, 
                                                                                                                                                   self.textcolor, fontsettings.font_name, self.textweight, int(self.maxvalue/2.0))
        else:
            y += self.rmax * 1.65
            svg += '<text text-anchor="middle" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%d</text>' % (x-0.01,y+fontsize/2-0.2, fontsize,
                                                                                                                                                    self.textcolor, fontsettings.font_name, self.textweight, int(self.maxvalue/2.0))
        svg += '<circle cx="%fvw" cy="%fvh" r="%fvh" style="fill:none; stroke:black; stroke-width:1.5; opacity:1.0;"></circle>' % (x,y, self.rmax/2.0)


        if sidetext:
            y += self.rmax * 1.0
            svg += '<text text-anchor="start" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%d</text>' % (x+self.rmax-0.3, y+fontsize/2-0.2, fontsize, 
                                                                                                                                                   self.textcolor, fontsettings.font_name, self.textweight, int(self.maxvalue/4.0))
        else:
            y += self.rmax * 0.85
            svg += '<text text-anchor="middle" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%d</text>' % (x-0.01,y+fontsize/2-0.2, fontsize,
                                                                                                                                                    self.textcolor, fontsettings.font_name, self.textweight, int(self.maxvalue/4.0))
        svg += '<circle cx="%fvw" cy="%fvh" r="%fvh" style="fill:none; stroke:black; stroke-width:1.5; opacity:1.0;"></circle>' % (x,y, self.rmax/4.0)
        
        svg += '</svg>'
        return svg
    
    
    # Return the SVG code
    def getsvg(self):
        svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0vw" y="0vh" width="%fvw" height="%fvh">' % (self.width, self.height)
        
        svg += '''
                <style type="text/css">
                   @import url('%s');
                   .fullrow:hover { stroke: #000044; stroke-dasharray: 3.5,3.5; cursor: pointer; !important; }
                </style>
               ''' % (fontsettings.font_url)

        
        # Title text
        if len(self.title) > 0:
            svg += '<text text-anchor="middle" x="%fvw"  y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % ((self.xstart+self.xend)/2.0, self.fontsize*1.75, self.titlefontsize, 
                                                                                                                                                   self.textcolor, fontsettings.font_name, self.textweight+100, self.title)
        
        # Text for the axis
        svg += verticalText(self.ycolumn, 0.4, self.ystart, 1.75*self.fontsize, self.textcolor, fontsettings.font_name, self.textweight+100)
        svg += '<text text-anchor="end" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (self.xstart, (self.fontsize + self.yend + 0.3), 1.75*self.fontsize, 
                                                                                                                                                self.textcolor, fontsettings.font_name, self.textweight+100, self.xcolumn)

        
        # Background
        svg += '<rect x="%fvw" y="%fvh" width="%fvw" height="%fvh" fill="%s" fill-opacity="1.0"></rect>' % (self.xstart, self.ystart, self.xend-self.xstart, self.yend-self.ystart, self.backcolor)
        
                
        # Vertical back lines and xaxis texts
        x = self.xstart
        for xvalue in self.xvalues:
            svg += '<line style="pointer-events:none; stroke:%s; stroke-width:0.2" x1="%fvw" y1="%fvh" x2="%fvw" y2="%fvh"></line>' % (self.backlinecolor, x, self.ystart, x, self.yend)
            if self.xtextangle == 0.0:
                if len(xvalue) <= self.maxlenx:
                    svg += '<text text-anchor="middle" x="%fvw"  y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (x + 0.5*self.xspace, self.yend+self.fontsize+0.3, self.fontsize, 
                                                                                                                                                         self.textcolor, fontsettings.font_name, self.textweight, xvalue)
                else:
                    rows = splitstring(xvalue,self.maxlenx)
                    y = 0
                    for r in rows:
                        svg += '<text text-anchor="middle" x="%fvw"  y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (x + 0.5*self.xspace, self.yend+self.fontsize+3+y, self.fontsize, 
                                                                                                                                                             self.textcolor, fontsettings.font_name, self.textweight, r)
                        y += self.fontsize
                    
            else:
                svg += '''
                        <svg x="%fvw" y="%fvh" overflow="visible">
                            <text style="pointer-events: none" text-anchor="end" x="0vw"  y="0vh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d" transform="rotate(%f)">%s</text>
                        </svg>
                ''' % (x + 0.5*self.xspace, self.yend+self.fontsize+0.3, self.fontsize, self.textcolor, fontsettings.font_name, self.textweight, self.xtextangle, xvalue)
                
            x += self.xspace
        svg += '<line style="pointer-events:none; stroke:%s; stroke-width:0.2" x1="%fvw" y1="%fvh" x2="%fvw" y2="%fvh"></line>' % (self.backlinecolor, x, self.ystart, x, self.yend)
            
            
        # Horizontal back lines and y axis texts
        y = self.ystart
        for yvalue in self.yvalues:
            svg += '<line style="pointer-events:none; stroke:%s; stroke-width:0.2" x1="%fvw" y1="%fvh" x2="%fvw" y2="%fvh"></line>' % (self.backlinecolor, self.xstart, y, self.xend, y)
            if len(yvalue) <= self.maxleny:
                svg += '<text text-anchor="end" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (self.xstart*0.92, y + 0.5*self.yspace, self.fontsize, 
                                                                                                                                                      self.textcolor, fontsettings.font_name, self.textweight, yvalue)
            else:
                rows = splitstring(yvalue,self.maxleny)
                dy = -self.fontsize*(len(rows)/2 - 1)
                for r in rows:
                    svg += '<text text-anchor="end" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (self.xstart*0.92, y + 0.5*self.yspace+dy, self.fontsize, 
                                                                                                                                                         self.textcolor, fontsettings.font_name, self.textweight, r)
                    dy += self.fontsize
                        
            y += self.yspace
        svg += '<line style="pointer-events:none; stroke:%s; stroke-width:0.2" x1="%fvw" y1="%fvh" x2="%fvw" y2="%fvh"></line>' % (self.backlinecolor, self.xstart, y, self.xend, y)


        # Cycle on all combinations of (x,y) values to display circles
        for xvalue in self.xvalues:
            for yvalue in self.yvalues:

                df = self.df[(self.df[self.xcolumn]==xvalue) & (self.df[self.ycolumn]==yvalue)].sort_values(by=[self.sizecolumn], ascending=False)
                
                # Cycle on circles sorted by descending dimension
                if df.shape[0] > 0:
                    ix = self.xvaluesindex[xvalue]
                    iy = self.yvaluesindex[yvalue]
                    x = self.xstart + self.xspace*(ix + 0.5)
                    y = self.ystart + self.yspace*(iy + 0.5)
                
                    dx = 0
                    dxnext = self.xspace/min(max(3.0,df.shape[0]), 8.0)
                    maxsize = df.iloc[0]['count']
                    maxr = self.getradius(maxsize)
                    for index, row in df.iterrows():
                        size = row[self.sizecolumn]
                        color = self.colorlist[0]
                        tooltip = xvalue + '/' + yvalue + ': ' + str(int(size))
                        
                        if self.mode == 'spread':
                            opacity = 0.7
                        else:
                            opacity = 1.0
                        
                        dodisplay = True
                        if self.ncolors > 0:
                            colorvalue = row[self.colorcolumn]
                            if colorvalue in self.colorvalues:
                                icolor = self.colorvalues.index(colorvalue)
                                color = self.colorlist[icolor % len(self.colorlist)]
                                if self.display[icolor]:
                                    tooltip = xvalue + '/' + yvalue + '/' + colorvalue + ': ' + str(int(size))
                                else:
                                    dodisplay = False

                        if dodisplay:
                            r = self.getradius(size)
                            yc = y
                            if self.mode == 'tangent': yc = y+maxr-r
                            svg += '<circle class="fullrow" cx="%fvw" cy="%fvh" r="%fvh" style="fill:%s; stroke:%s; stroke-width:%f; opacity:%f;"><title>%s</title></circle>' % (x+dx,yc, r, color, self.strokecolor,self.strokewidth, opacity, tooltip)

                        if self.mode == 'spread':
                            if dx == 0:  dx -= dxnext
                            elif dx < 0: dx = -dx
                            else:
                                dx = -dx - dxnext
                                dxnext = dxnext / 2.0

        
        # Display the legend
        self.legendpos = []

        y = self.yend + self.legenditemheight*1.1
        x = self.xstart
       
        svg += '<text text-anchor="start" x="%fvw"  y="%fvh" font-size="%fvh" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (x, y+self.fontsize*1.56, self.fontsize*1.75, 
                                                                                                                                                self.textcolor, fontsettings.font_name, self.textweight+100, self.colorcolumn)
        y = self.yend + 2*self.legenditemheight
        dx = self.legenditemwidth
        if self.ncolors > 0:
            maxlen = len(max(self.colorvalues, key=len))
            dx = max(maxlen*0.75,dx)
       
        for i in range(self.ncolors):
            icolor = i % len(self.colorlist)
            color = self.colorlist[icolor]
            name = self.colorvalues[i]

            if self.display[i]: opacity = 1.0
            else:               opacity = 0.3333
            svg += '<rect class="fullrow" style="cursor: pointer;" width="1.35vw" height="%fvh" x="%fvw" y="%fvh" fill="%s" fill-opacity="%f"><title>%s</title></rect>' % (2.7*self.legenditemheight/5, x, y, color, opacity, name)
            svg += '<text text-anchor="start" x="%fvw" y="%fvh" font-size="%fvh" fill="%s" fill-opacity="%f" style="font-family: %s;" font-weight="%d">%s</text>' % (x+1.6, y+self.fontsize*1.5, self.fontsize*1.5, 
                                                                                                                                                                  self.textcolor, opacity, fontsettings.font_name, self.textweight, name)
            
            self.legendpos.append([x,y])
                                  
            y += self.legenditemheight
            if (i+1) % self.legendrows == 0:
                y = self.yend + 2*self.legenditemheight
                x += dx

            
        svg += '</svg>'
        return svg

    
    # Returns Output widget containing the line chart
    def draw(self):
        self.out.clear_output(wait=True)
        with self.out:
            self.svg_picture = self.getsvg()
            display(HTML(self.svg_picture))
            #display(self.debug)

        return self.out
    
    




"""Dialog-box to display a message for the user."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import dialogGeneric
except:
    import dialogGeneric

#####################################################################################################################################################
# Display a dialog containing a simple message with a title
#####################################################################################################################################################
class dialogMessage(dialogGeneric.dialogGeneric):
    """
    Dialog-box to display a message for the user.
    
    Derived class from dialogGeneric.dialogGeneric.
        
    Parameters
    ----------
    title : str, optional
        Title of the dialog-box to be displayed in the top toolbar (default is '')
    text : str, optional
        Text to display on top of the dialog-box body (default is '')
    dark : bool, optional
        Flag that controls the color of the text in foreground (if True, the text will be displayed in white, elsewhere in black)
    show : bool, optional
        Flag to immediately show the dialog-box upon creation (default is False)
    width : int or str, optional
        Width of the dialog-box. If an integer is passed the width is intended in pixels. Default is 500 pixels
    addclosebuttons : bool, optional
        If True, the dialog will have a 'close' button in the top toolbar (default is True)
    transition : str, optional
        Transition to use for the dialog display and close (default is 'dialog-fade-transition'). See: https://vuetifyjs.com/en/styles/transitions/ for a list of available transitions (substitute 'v-' with 'dialog-')
    output : ipywidgets.Output, optional
        Output widget on which the widget has to be displayed
    titleheight : str, optional
        Height of the title toolbar. It can be: 'prominent', 'dense', 'extended' or a value in pixels (default is 'dense')
            
    Example
    -------
    Creation and display of a modal dialog-box containing an error message::
        
        from vois.vuetify import dialogMessage
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()
        display(output)

        e = dialogMessage.dialogMessage(title='Error',
                                        text='''Sorry but the task could not be completed<br>
        because there are errors in the code to save in PNG format''',
                                        addclosebuttons=False,
                                        show=True, width=450, output=output)
                              
    .. figure:: figures/dialogMessage.png
       :scale: 100 %
       :alt: dialogMessage widget

       Example of a dialogMessage to display an error message to the user.
   """
        
    def __init__(self, *args, **kwargs):
        
        text = kwargs['text'].replace('<br>','\n')
        vvv = text.split('\n')
        
        # Create the content to pass to the dialogGeneric
        if len(vvv) > 0:
            ctexts = [ v.CardText(children=[vvv[0]], class_="mt-1") ] + [ v.CardText(children=[x], class_="mt-n8") for x in vvv[1:] ]
            kwargs['content'] = [v.Card(children=ctexts)]
            
        kwargs['text'] = ''
        kwargs['fullscreen'] = False
        #kwargs['addclosebuttons'] = True
        
        super().__init__(*args, **kwargs)




"""Side panel that opens on the side of the screen to show content or get user input."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
    from . import tooltip
except:
    import settings
    import tooltip


#####################################################################################################################################################
# Display a side-panel with a close button
#####################################################################################################################################################
class sidePanel():
    """
    Side panel that opens on the side of the screen to show content or get user input.
        
    Parameters
    ----------
    title : str, optional
        Title to display on top of the side panel (default is '')
    text : str, optional
        String of text to display in the content of the side panel (default is '')
    width : int, optional
        Width of the panel in pixel (default is 500 pixels)
    right : bool, optional
        Flag that controls if the panel is opened on the right side of the screen (default is False)
    zindex : int, optional
        Z-index assigned to the panel (default is 9999)
    showclosebuttons : bool, optional
        Flag to display the close button in the top bar of the side panel (default is True)
    dark : bool, optional
        Flag that controls the color of the text in foreground (if True, the text will be displayed in white, elsewhere in black)
    backdark : bool, optional
        Flag that controls the background color of the panel (if True, the background will be black, elsewhere white)
    content : list of widgets, optional
        Additional content to be added to the side panel to get user input (default is [])
    output : ipywidgets.Output, optional
        Output widget on which the side panel has to be displayed
    onclose : function, optional
        Python function to call when the user closes the side panel. The function will be called withput any parameter
            
    Example
    -------
    Creation and display of a side panel::
        
        from vois.vuetify import sidePanel, slider
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()
        display(output)

        def onclose():
            with output:
                print('CLOSED')

        text  = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore...'

        s = slider.slider(66,0,100)

        panel = sidePanel.sidePanel(title='Help panel', text=text,
                                    width=400, content=[s.draw()],
                                    output=output, onclose=onclose,
                                    dark=False)
        panel.show()

    .. figure:: figures/sidepanel.png
       :scale: 100 %
       :alt: sidepanel widget

       Example of a side panel containing text and a slider widget.
   """

    def __init__(self, title='', text='', width=500, right=False, zindex=9999, showclosebuttons=True, dark=settings.dark_mode, backdark=settings.dark_mode, content=[], output=None, onclose=None):
        
        self.onclose = onclose
        
        text = text.replace('<br>','\n')
        vvv = text.split('\n')

        if dark:
            buttontext = settings.textcolor_dark
            styletext = 'color: ' + settings.textcolor_dark + ';'
        else:
            buttontext = settings.textcolor_notdark
            styletext = 'color: ' + settings.textcolor_notdark + ';'

        ttitle  = v.ToolbarTitle(children=[title], style_=styletext)
        spacer  = v.Spacer()

        if showclosebuttons:
            # Close button
            bclose = v.Btn(text=True, children=['Close'], color=buttontext)
            bclose.on_event('click', self.close)
            bclose = tooltip.tooltip('Close the panel',bclose)

            # Close X button
            bx = v.Btn(icon=True, children=[v.Icon(children=['mdi-close'])], color=buttontext)
            bx.on_event('click', self.close)
            bx = tooltip.tooltip('Close the panel',bx)

            # Toolbar
            titems  = v.ToolbarItems(children=[bclose])
            toolbar = v.Toolbar(dark=True, color=settings.color_first, children=[bx,spacer,ttitle,spacer,titems])
        else:
            toolbar = v.Toolbar(dark=True, color=settings.color_first, children=[spacer,ttitle,spacer])
        
        # Dialog
        backcolor = 'white'
        if backdark: backcolor = settings.dark_background
        self.nav = v.NavigationDrawer(value=False, app=True, color=backcolor, width=width, style_='z-index:%d;' % zindex, right=right,
                                      children=[v.Card(elevation=0, children=[ toolbar, v.CardText(class_='pa-0 ma-0 mt-9', children=['']) ] + 
                                                                [ v.CardText(children=[x], class_="mt-n5") for x in vvv ] + content
                                                      )
                                               ])

        # Display of the panel
        if not output is None:
            with output:
                display(self.nav)
        
    # Open the panel
    def show(self):
        """Displays the side panel widget"""
        self.nav.value = True

    # Close the panel
    def close(self,*args):
        """Closes the side panel widget"""
        self.nav.value = False
        if self.onclose:
            self.onclose()
        
    # Returns True if the panel is open, False otherwise
    def isopen(self):
        """Queries the status of the side panel widget: returns True if the panel is displayed, False if it was closed"""
        return self.nav.value


"""Map popup widget to display titles and texts in a geographic position on a ipyleaflet Map."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Widgets import
import ipyleaflet
from ipyleaflet import AwesomeIcon, DivIcon, Marker

# Python imports
from threading import Timer


#####################################################################################################################################################
# textpopup class
#####################################################################################################################################################
class textpopup():
    """
    Widget to vertically display a list of titles and texts strings as a popup on a ipyleaflet Map. Each couple of title and text occupies a row.
        
    Parameters
    ----------
    m : instance of ipyleaflet.Map class
        Map to which the popup has to be added
    lat : float
        Latitude where the popup has to be displayed (default is 0.0)
    lon : float
        Longitude where the popup has to be displayed (default is 0.0)
    titles : list of strings
        Strings to be displayed as title of each row (default is [])
    texts : list of strings
        Strings to be displayed as the content of each row (default is [])
    width : int
        Width of the popup in pixels (default is 200)
    height : int
        Height of the popup in pixel (default is None meaning that the height will be automatically calcolated)
    autoremovedelay : float
        Time in seconds for automatic remnoving of the popup (default is 0.0 which disables autoremove)
    titlesbold : list of strings, optional
        List of titles whose corresponding texts in the left column should be displayed using bold font (default is [])
    titlefontsize : int, optional
        Size in pixel of the font used for the titles (default is 12)
    textsbold : list of strings, optional
        List of titles whose corresponding texts in the right column should be displayed using bold font (default is [])
    textfontsize : int, optional
        Size in pixel of the font used for the texts (default is 12)
    titlecolor : str, optional
        Color to use for the titles (default is 'black')
    textcolor : str, optional
        Color to use for the texts (default is 'black')
    lineheightfactor : float, optional
        Factor to multiply to the font-size to calculate the height of each row (default is 1.5)

    Example
    -------
    Creation and display of a widget to display some textual information::
        
        from ipywidgets import widgets, HTML, CallbackDispatcher
        from ipyleaflet import Map
        from IPython.display import display

        from vois import textpopup

        m = Map(center=[43.66737, 12.5504], scroll_wheel_zoom=True, zoom=13)

        t = None
        def handle_interaction_popup(**kwargs):
            global t

            if kwargs.get('type') == 'click':
                lat = kwargs.get('coordinates')[0]
                lon = kwargs.get('coordinates')[1]

                textpopup.textpopup.removeAll(m)
                t = textpopup.textpopup(m, lat=lat, lon=lon, autoremovedelay=5.0,
                                        width=340, height=None, titlewidth=70,
                                        titles=['Pixel values', 'Class'],
                                        texts=['(120,34,189)', 'Woodland and Shrubland (incl. permanent crops)'],
                                        titlesbold=[],
                                        titlefontsize=11,
                                        textsbold=['Pixel'],
                                        textfontsize=11,
                                        titlecolor='darkgreen',
                                        textcolor='darkred')

        m._interaction_callbacks = CallbackDispatcher()
        m.on_interaction(handle_interaction_popup)

        display(m)

    .. figure:: figures/textpopup.png
       :scale: 100 %
       :alt: textpopup widget

       Map popup widget for displaying textual information.
   """
    
    def __init__(self,
                 m,
                 lat=0.0,
                 lon=0.0,
                 titles=[],
                 texts=[],
                 width=200,
                 height=None,
                 autoremovedelay=10.0,
                 titlesbold=[],
                 titlefontsize=12,
                 textsbold=[],
                 textfontsize=12,
                 titlewidth=50,
                 titlecolor='black',
                 textcolor='black',
                 lineheightfactor=1.1
                ):
        self.m = m
        self.autoremovedelay = autoremovedelay
    
        lineheight = "line-height: %dpx;"%(int(lineheightfactor*(max(titlefontsize,textfontsize)))) # To ensure vertical center alignment
    
        # Autocalc height
        if height is None:
            height = max(titlefontsize,textfontsize)*lineheightfactor*1.1*max(len(titles),len(texts)) + 12
    
        margin = 5
        width  += 2*margin
        height += margin
        
        self.h = '<table border="0" style="border-collapse: collapse; margin-right: %dpx; margin-left: %dpx; margin-top: %dpx; margin-bottom: 0px; display: block; width: %dpx; height: %dpx; overflow-y: auto;"><tbody>'%(margin,margin,margin, width,height)
        for i, title in enumerate(titles):
            if i < len(texts):
                text = texts[i]
            else:
                text = ''
            if title in titlesbold: tdtitle = 'th'
            else:                   tdtitle = 'td'
            if title in textsbold:  tdtext  = 'th'
            else:                   tdtext  = 'td'
            self.h += '''<tr style="border-bottom: 1px solid lightgrey;">
  <%s align="center" style="width: %dpx; font-size: %dpx; color: %s; %s">%s</%s>
  <%s align="center" style="width: %dpx; font-size: %dpx; color: %s; %s">%s</%s>
</tr>'''%(tdtitle,titlewidth,titlefontsize,titlecolor,lineheight,title,tdtitle, 
          tdtext,width-titlewidth-4*margin,textfontsize,textcolor,lineheight,text,tdtext)

        self.h += '</tbody></table>'
        
        center = (lat,lon)
        icon1 = AwesomeIcon(name='', marker_color='white', icon_color='white', spin=False)
        self.marker1 = Marker(name='textpopup', icon=icon1, location=center)

        icon2 = DivIcon(html=self.h, icon_anchor=[width/2, height+14], icon_size=[width, height])
        self.marker2 = Marker(name='textpopup', location=center, icon=icon2)
    
        self.m.add_layer(self.marker1)
        self.m.add_layer(self.marker2)
        
        # Auto-remove after some time
        if self.autoremovedelay > 0:
            self.timer = Timer(self.autoremovedelay, self.remove)
            self.timer.start()
    
    
    # Remove the textpopup from the map
    def remove(self):
        if self.autoremovedelay > 0:
            self.timer.cancel()
            
        if self.marker1 in self.m.layers:
            self.m.remove_layer(self.marker1)
        if self.marker2 in self.m.layers:
            self.m.remove_layer(self.marker2)
        
    
    # Remove all textpopups from a map
    @staticmethod
    def removeAll(m):
        for layer in reversed(m.layers):
            if isinstance(layer, ipyleaflet.leaflet.Marker) and layer.name == 'textpopup':
                m.remove_layer(layer)
